"""
This DIT module is part of DuIvyTools library. Written by CharlesHahn.

DIT module provides simple API and CLI commands for you to use conveniently. 

This file is provided to you under GPLv3 License.
"""


import os
import sys
import logging
import matplotlib.pyplot as plt

base = os.path.dirname(os.path.realpath(__file__))
if base not in sys.path:
    sys.path.insert(0, base)

from XPM import xpm_call_functions
from XVG import xvg_call_functions
from NDX import ndx_call_functions
from MDP import mdp_call_functions
from HELP import help_call_functions, get_welcome_msg
from PipiDistAng import pipi_dist_ang_call_functions
from FindCenter import find_center_call_functions
from HydrogenBond import hbond_call_functions
from MolMap import mol_map_call_functions
from DCCM import dccm_call_functions
from DSSP import dssp_call_functions
from MplStyle import mplstyle_call_functions


logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
logger = logging.getLogger(__name__)


def load_style():
    style_files = [file for file in os.listdir() if file[-9:] == ".mplstyle"]
    if len(style_files) >= 1:
        plt.style.use(style_files[0])
        logging.info("using matplotlib style sheet from {}".format(style_files[0]))
    else:
        data_file_path = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__))
        )
        mplstyle = os.path.join(data_file_path, os.path.join("data", "DIT.mplstyle"))
        plt.style.use(mplstyle)
        logging.info("using default matplotlib style sheet")


def main():
    arguments = [argv for argv in sys.argv]
    if len(sys.argv) < 2:
        DIT_infos = get_welcome_msg()
        print(DIT_infos)
        sys.exit()
    elif len(sys.argv) == 2:
        if sys.argv[1] in ["help", "-h", "--help"]:
            help_call_functions(arguments)
        elif sys.argv[1] == "show_style":
            mplstyle_call_functions(arguments)
        else:
            logging.error("unknown command {}, ".format(sys.argv[1]))
            logging.info("type `dit help` for more information")
            sys.exit()
    elif len(sys.argv) == 3:
        if sys.argv[1] == "help":
            help_call_functions(arguments)
        elif sys.argv[2] in ["-h", "--help", "help"]:
            help_call_functions(["dit", "help", sys.argv[1]])
        else:
            logging.error("wrong command, type `dit help` for more information")
            sys.exit()
    elif len(sys.argv) > 3:
        load_style()  # to load style here
        method = sys.argv[1]
        if method.startswith("xvg"):
            xvg_call_functions(arguments)
        elif method.startswith("xpm"):
            xpm_call_functions(arguments)
        elif method.startswith("ndx"):
            ndx_call_functions(arguments)
        elif method.startswith("mdp"):
            mdp_call_functions(arguments)
        elif method.startswith("help"):
            help_call_functions(arguments)
        elif method == "pipi_dist_ang":
            pipi_dist_ang_call_functions(arguments)
        elif method == "find_center":
            find_center_call_functions(arguments)
        elif method == "hbond":
            hbond_call_functions(arguments)
        elif method == "mol_map":
            mol_map_call_functions(arguments)
        elif method == "dccm_ascii":
            dccm_call_functions(arguments)
        elif method == "dssp":
            dssp_call_functions(arguments)
        else:
            logging.error("unknown command {}".format(method))


if __name__ == "__main__":
    main()



"""
HELP module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to provide help messages. 

This HELP module contains:
    - HELP class
    - help_call_functions

This file is provided to you by GPLv3 license.
"""

import sys
import logging


class HELP(object):
    """HELP class provides help infos about all method in DuIvyTools"""

    def __init__(self) -> None:
        """init all help messages"""

        self.help_infos = {
            "xvg_show": """
xvg_show: draw xvg data into line figures.

:examples: 
    dit xvg_show -f test.xvg
    dit xvg_show -f test.xvg --subplot
    dit xvg_show -f test.xvg -o test.png -ns

:parameters:
    -f, --input
            specify the xvg file for input.
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
    -subplot, --subplot (optional)
            whether to draw different data in subplots.
""",
            "xvg_compare": """
xvg_compare: comparison of xvg files, draw different data columns you selected
             into line figure.

:examples:
    dit xvg_compare -f f1.xvg f2.xvg -c 1,2 2,3,4 -l l1 l2 l3 l4 l5
    dit xvg_compare -f f1.xvg f2.xvg -c 1 1 -l l1 l2 -x Time(ns) -y ylabel 
    dit xvg_compare -f f1.xvg f2.xvg -c 1,2 2 -l l1 l2 l3 -s 100 -e 500 -t test
    dit xvg_compare -f f1.xvg f2.xvg -c 1,2 2,3,4 -o test.png -ns

:parameters:
    -f, --input
            specify the xvg files for input. accept one or one more input file
            names seperated by space.
    -c, --column_select
            specify the index of columns you want to use. index starts from 0. 
            one xvg file in -f should be paired with one column index group 
            seperated by space, and in each column index group, indexs should 
            be seperated by comma. 
            In first example, 1,2 is paired with f1.xvg, 2,3,4 is paird with 
            f2.xvg. This means selecting the column 1 and 2 in f1.xvg, column 
            2, 3 and 4 in f2.xvg to compare. 
    -l, --legend_list (optional) 
            specify the legends to show in figure. The number of legends should
            be equal to the number of indexs you sepecified in -c. if you want 
            space in one legend, like "hh gg"; use "~~" for instead, 
            like "hh~~gg", ~~ will be turn into space and shown in figure.
    -s, --start (optional)
            specify row index of column data which you want to start to plot.
    -e, --end (optional)
            specify row index of column data which you want to stop plotting.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
    -t, --title (optional)
            specify the title you want to show in figure.
    -smv, --showMV  (optional)
            if you specified this parameter, moving average and confidence 
            interval will be calculated and shown in figure. 
    -ws, --windowsize (optional)
            specify the windows size for moving average calculation. 
            default == 50, only valid when -smv specified. 
    -cf, --confidence (optional)
            specify the confidence for confidence interval calculation. 
            default == 0.90, only valid when -smv specified. 
    -a, --alpha (optional)
            specify the transparency for plotting confidence interval. 
            default == 0.4, only valid when -smv specified. 
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xvg
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
""",
            "xvg_ave": """
xvg_ave: calculate the average of each column in xvg file.

:examples:
    dit xvg_ave -f f1.xvg
    dit xvg_ave -f f1.xvg -s 1000 -e 2000

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -s, --start (optional)
            specify the row index of column data which you want to start to 
            calculation.
    -e, --end (optional)
            specify the row index of column data which you want to stop 
            calculation.
""",
            "xvg_mvave": """
xvg_mvave: calculate the moving average of each column in xvg file and save 
           results to csv file.

:examples:
    dit xvg_mvave -f f1.xvg -o res.csv 
    dit xvg_mvave -f f1.xvg -o res.csv -ws 20 -cf 0.95

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -o, --output
            specify the csv file to save results.
    -ws, --windowsize (optional)
            specify the windows size for moving average calculation. 
            default == 50
    -cf, --confidence (optional)
            specify the confidence for confidence interval calculation. 
            default == 0.90
""",
            "xvg2csv": """
xvg2csv: convert xvg file into csv file.

:examples:
    dit xvg2csv -f f1.xvg -o res.csv

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -o, --output
            specify the csv file to save results.
""",
            "xvg_rama": """
xvg_rama: draw ramachandran figure from a rama data xvg file.

:examples:
    dit xvg_rama -f rama.xvg
    dit xvg_rama -f rama.xvg -o test.png -ns

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
""",
            "xvg_show_distribution": """
xvg_show_distribution: calculate distribution of each data column in xvg file 
                       and show.

:examples:
    dit xvg_show_distribution -f f1.xvg
    dit xvg_show_distribution -f f1.xvg -bin 100
    dit xvg_show_distribution -f f1.xvg -o test.png -ns

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -bin, --bin (optional)
            specify the number of bins of calculating distribution. 
            default == 100
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
""",
            "xvg_show_stack": """
xvg_show_stack: draw stack figure of column data.

:examples:
    dit xvg_show_stack -f f1.xvg -c 2 3 4 5 6
    dit xvg_show_stack -f f1.xvg -c 2 3 4 5 6 -l A B C D E -s 500 -e 1000
    dit xvg_show_stack -f f1.xvg -c 2 3 4 5 6 -o test.png -ns

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -c, --column_select
            specify the column indexs you want to stack,  seperated by space.
    -l, --legend_list (optional)
            Normally, legends could be assigned automatically, but you can 
            specify them by -l. The number of legends should be equal to the 
            number of column indexs. if you want space in one legend, like 
            "hh gg", use "~~" for instead, like "hh~~gg". ~~ will be turn into 
            space and shown in figure.
    -s, --start (optional)
            specify the row index of column data which you want to start to 
            calculation.
    -e, --end (optional)
            specify the row index of column data which you want to stop 
            calculation.
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xvg
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
""",
            "xvg_show_scatter": """
xvg_show_scatter: draw scatter figure.

:examples:
    dit xvg_show_scatter -f f1.xvg -xi 0 -yi 1
    dit xvg_show_scatter -f f1.xvg -xi 1 -yi 2 -c 0
    dit xvg_show_scatter -f f1.xvg -xi 0 -yi 1 -o test.png -ns

:parameters:
    -f, --input
            specify the xvg file for input. 
            Only the first file is valid, others will be dropped.
    -xi, --x_index (optional)
            specify the index of column which you want to use as x data. 
            default == 0
    -yi, --y_index (optional)
            specify the index of column which you want to use as y data. 
            default == 1
    -c, --column_select (optional)
            specify the index of column which you want to use as color 
            mapping, select only one index
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
""",
            "xvg_energy_compute": """
xvg_energy_compute: calculate the interaction energy between two items.

:examples:
    dit xvg_energy_compute -f prolig.xvg pro.xvg lig.xvg -o energy_results.xvg

:parameters:
    -f, --input
            specify three xvg files to calculate (complex energy file, 
            component A energy file and component B energy file). Each should 
            contain and ONLY contain five columns 
            (Time, LJ(SR), Disper.corr., Coulomb(SR), Coul.recip.).
    -o, --output
            specify a xvg file to save results. 
""",
            "xvg_combine": """
xvg_combine: combine data columns selected and save results into one xvg file. 

:examples:
    dit xvg_combine -f f1.xvg -c 1,2 -o res.xvg
    dit xvg_combine -f f1.xvg f2.xvg -c 1,2 2,3 -o res.xvg

:parameters:
    -f, --input
            specify the xvg files for input. accept one or one more input file
            names seperated by space.
    -c, --column_select
            specify the index of columns you want to use. index starts from 0. 
            one xvg file in -f should be paired with one column index group 
            seperated by space. And in each column index group, indexs should
            be seperated by comma. 
            In second example, 1,2 is paired with f1.xvg, 2,3 is paird with 
            f2.xvg. This means selecting the column 1 and 2 in f1.xvg, column 2
            and 2 in f2.xvg to combine, and saving results to res.xvg. 
    -o, --output
            specify a xvg file to save results of combination.
""",
            "xvg_ave_bar": """
xvg_ave_bar: First, the average of each data column you select by -c will be
             calculated.
             Second, the average and standard error of corresponding column 
             averages in xvg file groups will be calculated.
             Third, the averages and standard errors will be drawn into bar 
             figure.

:examples:
    dit xvg_ave_bar -f f1_1.xvg,f1_2.xvg,f1_3.xvg f2_1.xvg,f2_2.xvg,f2_3.xvg -c 1 2 3
    dit xvg_ave_bar -f f1.xvg f2.xvg -c 1 2 -l L1 L2 -s 100 -e 1000 
    dit xvg_ave_bar -f f1.xvg f2.xvg -c 1 2 -x xlabel -y ylabel -t test
    dit xvg_ave_bar -f f1.xvg f2.xvg f3.xvg -c 1 2 -l L1 L2 L3 -xt xt1 xt2
    dit xvg_ave_bar -f f1.xvg f2.xvg -c 1 2 -l L1 L2 -ac average.csv
    dit xvg_ave_bar -f f1.xvg f2.xvg -c 1 2 -l L1 L2 -o test.png -ns

:parameters:
    -f, --input
            specify the xvg files for input. accept one or one more file names
            or file name groups. 
            Several file names seperated by comma is called a file name group. 
            In first example, f1_1.xvg,f1_2.xvg,f1_3.xvg is the first file name
            group, and f2_1.xvg,f2_2.xvg,f2_3.xvg is second file name group.
            This method will calculate averages of columns you select in each
            xvg file. Then calculate the average and standard error of column
            averages in each file group. Finally, convert the averages and 
            standard errors into bar plot.
    -c, --column_select
            specify the index of columns you want to calculate. Index starts 
            from 0. The column indexs you select will be applied to EVERY xvg
            file in -f.
    -l, --legend_list (optional) 
            specify the legends to show in figure. The number of legends should
            be equal to the number of files or file groups you specified in -f.
            If you want space in legend, like "hh gg"; use "~~" for instead, 
            like "hh~~gg", ~~ will be turn into space and shown in figure.
    -xt, --xtitles (optional)
            specity the names of x ticks in figure. The number of xtitles 
            should be equal to the number of indexs in column_select (-c).
    -s, --start (optional)
            specify the row index of column data which you want to start to 
            calculation.
    -e, --end (optional)
            specify the row index of column data which you want to stop 
            calculation.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
    -t, --title (optional)
            specify the title you want to show in figure.
    -ac, --ave2csv (optional)
            the csv file name for saving data of averages into csv file.
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
""",
            "xvg_box": """
xvg_box: draw the data of columns you select into box figure.

:examples:
    dit xvg_box -f f1.xvg f2.xvg f3.xvg -c 1 2 -xt xt1 xt2 -s 100 -e 1000
    dit xvg_box -f f1.xvg f2.xvg -c 1 2 3 -xt xt1 xt2 xt3 -x xlabel -y ylabel
    dit xvg_box -f f1.xvg f2.xvg f3.xvg -c 1 2 -o test.png -ns -t title

:parameters:
    -f, --input
            accept one or more than one files as input.
    -c, --column_select
            specify the column indexs you want to plot. The indexs you 
            specified will be applied to each file you select in -f.
    -xt, --xtitles (optional)
            specity the names of x ticks in figure. The number of xtitles 
            should be equal to the number of indexs in column_select (-c).
    -s, --start (optional)
            specify the row index of column data which you want to start to 
            calculation.
    -e, --end (optional)
            specify the row index of column data which you want to stop 
            calculation.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
    -t, --title (optional)
            specify the title you want to show in figure.
    -o, --output (optional)
            specify the output figure name. Figure will be saved in 300 dpi.
    -ns, --noshow (optional)
            if -ns is specified, figure won't be shown on screen. Usually work
            with -o to save figure directly.
""",
            "xpm_show": """
xpm_show: vasualize the xpm file.

:examples:
    dit xpm_show -f f1.xpm 
    dit xpm_show -f f1.xpm -ip 
    dit xpm_show -f f1.xpm -pcm 
    dit xpm_show -f f1.xpm -3d
    dit xpm_show -f f1.xpm -o test.png -ns 
    dit xpm_show -f 31.xpm -x xlabel -y ylabel -t title
    dit xpm_show -f 31.xpm -x xlabel -xs 0.001


:parameters:
    -f, --input
            specify the xpm file for input. 
    -o, --output (optional)
            specify the output figure name. 
    -ip, --interpolation (optional)
            whether to perform interpolation which will make your figure looks 
            smoother.
    -pcm, --pcolormesh (optional)
            whether to use pcolormesh function to vasualize your xpm figure.
    -3d, --threeDimensions (optional)
            whether to draw a 3D figure. Useful for free energy landscape plot.
    -ns, --noshow (optional)
            whether not to show figure in GUI, useful when working without GUI.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
    -t, --title (optional)
            specify the title you want to show in figure.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xpm
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
""",
            "xpm2csv": """
xpm2csv: convert xpm file into csv data file. 
         when number of x-axis (or y-axis) is 1 more than width (or height) of
         figure, the intermediate value of two adjacent values of x-axis (or
         y-axis) will be calculated to suit the width (or height).

:examples:
    dit xpm2csv -f f1.xpm -o test.csv
    dit xpm2csv -f f1.xpm -o test.csv -x xlabel -y ylabel
    dit xpm2csv -f f1.xpm -o test.csv -x xlabel -xs 0.001

:parameters:
    -f, --input
            specify the xpm file for input. 
    -o, --output
            specify the output csv file name.
    -x, --xlabel (optional)
            specify the x-label you want to write to csv file.
    -y, --ylabel (optional)
            specify the y-label you want to write to csv file.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xpm
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
""",
            "xpm2dat": """
xpm2dat: convert xpm file into N*N data file. 
         when number of x-axis (or y-axis) is 1 more than width (or height) of
         figure, the intermediate value of two adjacent values of x-axis (or
         y-axis) will be calculated to suit the width (or height).

:examples:
    dit xpm2dat -f f1.xpm -o test.dat
    dit xpm2dat -f f1.xpm -o test.dat -x xlabel -y ylabel
    dit xpm2dat -f f1.xpm -o test.dat -x xlabel -xs 0.001

:parameters:
    -f, --input
            specify the xpm file for input. 
    -o, --output
            specify the output dat file name.
    -x, --xlabel (optional)
            specify the x-label you want to write to csv file.
    -y, --ylabel (optional)
            specify the y-label you want to write to csv file.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xpm
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
""",
            "xpm2gpl": """
xpm2gpl: convert xpm file into gnuplot script.

:examples:
    dit xpm2gpl -f f1.xpm -o test.gpl
    dit xpm2gpl -f f1.xpm -o test.gpl -x xlabel -y ylabel -t title
    dit xpm2gpl -f f1.xpm -o test.gpl -x xlabel -xs 0.001

:parameters:
    -f, --input
            specify the xpm file for input. 
    -o, --output
            specify the output gnuplot script file name.
    -x, --xlabel (optional)
            specify the x-label you want to show in figure.
    -y, --ylabel (optional)
            specify the y-label you want to show in figure.
    -t, --title (optional)
            specify the title you want to show in figure.
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xpm
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
            I advise you'd better to use -tu in gmx commands for convertion of 
            unit of x-axis.
""",
            "xpm_combine": """
xpm_combine: combine several xpm files into one figure. 
             STILL WORKING ON IT. NOT RECOMMAND TO USE IT !!!

:examples:
    dit xpm_combine -c f1.xpm f2.xpm
    dit xpm_combine -c f1.xpm f2.xpm -o test.png -ns

:parameters:
    -c, --combine
            specify several xpm files for input. 
    -o, --output
            specify the output figure name. 
    -ns, --noshow (optional)
            whether not to show figure in GUI, useful when working without GUI.
""",
            "ndx_show": """
ndx_show: print all group names in ndx file you specified.

:examples:
    dit ndx_show -f f1.ndx

:parameters:
    -f, --input
            specify the ndx file for input. 
""",
            "ndx_rm_dup": """
ndx_rm_dup: remove all duplicated groups in ndx file.

:examples:
    dit ndx_rm_dup -f f1.ndx -o res.ndx

:parameters:
    -f, --input
            specify the ndx file for input. 
    -o, --output
            specify the ndx file name for saving the results of romiving.
""",
            "ndx_rm": """
ndx_rm: remove index groups according to names you specified.

:examples:
    dit ndx_rm -f f1.ndx -o res.ndx -gl Protein SOL System
    dit ndx_rm -f f1.ndx -o res.ndx -int

:parameters:
    -f, --input
            specify the ndx file for input. 
    -o, --output
            specify the ndx file name for saving the results.
    -gl, --grouplist (optional)
            specify the group names you want to remove. Seperated by space.
    -int, --interactive (optional)
            specify the group names you want to remove by interactive mode.
""",
            "ndx_preserve": """
ndx_preserve: preserve index groups according to names you specified, and 
              remove others.

:examples:
    dit ndx_preserve -f f1.ndx -o res.ndx -gl Protein SOL System
    dit ndx_preserve -f f1.ndx -o res.ndx -int

:parameters:
    -f, --input
            specify the ndx file for input. 
    -o, --output
            specify the ndx file name for saving the results.
    -gl, --grouplist (optional)
            specify the group names you want to preserve. Seperated by space.
    -int, --interactive (optional)
            specify the group names you want to preserve by interactive mode.
""",
            "ndx_add": """
ndx_add: add one index group to ndx file by specified group name(-gn), 
         start(-s), end(-e), step(-t).

:examples:
    dit ndx_add -f f1.ndx -o res.ndx -gn test -s 1 -e 10 -t 2

:parameters:
    -f, --input (optional)
            specify ndx file for input. -f could be ignored if you want a new
            ndx file with the index group added.
    -o, --output
            specify the ndx file name for saving the results.
    -gn, --groupname
            specify the name of index group you want to add.
    -s, --start
            specify the start index of the group you want to add.
    -e, --end
            specify the end index of the group you want to add.
    -t, --step (optional)
            specify the step of indexs of the group you want to add. 
            default == 1. In the example, -s 1 -e 10 -t 2 means to generate a
            group contains index: 1 3 5 7 9 
""",
            "ndx_combine": """
ndx_combine: combine groups into one group by the group names you specified.
             I recommand you to use gmx make_ndx command to do this.

:example:
    dit ndx_combine -f f1.ndx -o res.ndx -gn combined_group -gl Protein Ligand

:parameters:
    -f, --input
            specify the ndx file for input. 
    -o, --output
            specify the ndx file name for saving the results.
    -gn, --groupname
            specify the name of combined index group.
    -gl, --grouplist
            specify the group names you want to combine. Seperated by space.
""",
            "ndx_rename": """
ndx_rename: rename groups in ndx file.

:examples:
    dit ndx_rename -f f1.ndx -o res.ndx -on Protein -nn test
    dit ndx_rename -f f1.ndx -o res.ndx -int

:parameters:
    -f, --input
            specify the ndx file for input. 
    -o, --output
            specify the ndx file name for saving the results.
    -on, --oldname (optional)
            specify the old name of group you want to rename.
    -nn, --newname (optional)
            specify a new name for the group you renamed.
    -int, --interactive (optional)
            rename groups in interactive mode.
""",
            "mdp_gen": """
mdp_gen: generate a template mdp file by application you specified. The mdp 
         file generated may NOT be appropriate for your system. 
         CHECK IT YOURSELF !

:examples:
    dit mdp_gen -o em.mdp -a em
    dit mdp_gen -o npt.mdp -a npt

:parameters:
    -o, --output
            specify an output mdp file to save the results.
    -a, --application
            specify the applications you want to use mdp file for. You can 
            select from: ions, em, nvt, npt, md, blank
""",
            "pipi_dist_ang": """
pipi_dist_ang: to calculate the distance and angles between two rings or between 
               one ring and one vector defined by atom group or command line.

:examples:
    dit pipi_dist_ang -f test.gro -n test.ndx 
    dit pipi_dist_ang -f test.gro -n test.ndx -select ring1 ring2
    dit pipi_dist_ang -f test.gro -n test.ndx -select ring1 -vg
    dit pipi_dist_ang -f test.gro -n test.ndx -select ring1 -vec 6 6 6 

:parameters:
    -f, --input
            the gro file which contains frames of molecule coordinates
    -n, --index
            the index file which contains atom index groups of rings or 
            vector groups
    -b (optional)
            the frame number to start calculation, default=0
    -dt (optional)
            the frame interval, default=1
    -o, --output
            the output filename to save results
    -vg (optional)
            whether to get vector from index group, default=False
    -vec (optional)
            specify the vector by command line, eg. -vec 6 6 6 
    -select (optional)
            select the groups from command line
""",
            "find_center": """
find_center: to find the atom which is nearest to center of atom group. 

:examples:
    dit find_center -f test.gro
    dit find_center -f test.gro -n index.ndx
    dit find_center -f test.gro -n index.ndx -aa

:parameters:
    -f, --input
            the gro file which contains one frame of molecule coordinates
    -n, --index (optional)
            the index file which you could select group from
    -aa, --AllAtoms (optional)
            if to find center of one group atoms in all atoms
""",
            "hbond": """
hbond: A useful method to process hbond related files generated by `gmx hbond`,
        you can get hbond occupancy figure and table infos from `dit hbond`. 
        note: the hbond names will show on figure only when number of hbonds 
              less than 10. 

:examples:
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -c 5 7 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -c 5 7 0-3
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -o test.png -csv test.csv -ns
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf d_resname(d_resnum)@d_atomname(d_atomnum)->h_atomname(h_atomnum)...a_resname(a_resnum)@a_atomname(a_atomnum) 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf "d_resname@d_atomname -> h_atomname ... a_resname@a_atomname" 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf "d_atomname@h_atomname -> a_atomname" 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf "number" 
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf id -so AND0,2,5-7
    dit hbond -f md.gro -n hbond.ndx -m hbmap.xpm -hnf id -so OR0,2,5-7

:parameters:
    -f, --input
            the gro file which contains one frame of molecule coordinates, 
            used to get atom names.
    -n, --index
            the index file generated by `gmx hbond`
    -m, --map
            the hbond map file (hbmap.xpm) generated by `gmx hbond`
    -c, --select (optional)
            select the hbond id to only show the selected hbonds
    -o, --output (optional)
            the figure name you want to save
    -csv, --csv (optional)
            save the table infos into a csv whose name is specified here
    -ns, --noshow (optional)
            whether not to show figure, useful on computer without gui
    -hnf, --hbond_name_format (optional)
            define the hbond name format by user! 
            Each atom has four features: resname, resnum, atomname, atomnum. 
            Distinguish donor, hydrogen, acceptor by adding one prefix to each
            feature, like: d_resname, a_resnum, h_atomname. So you may able to
            define hbond name style by: 
            'd_resname(d_resnum)@d_atomname(d_atomnum)->h_atomname(h_atomnum)...a_resname(a_resnum)@a_atomname(a_atomnum)' 
            which is the default style, or also you could specify 
            'd_atomname@h_atomname...a_atomname' 
            or some format you would like.
            Or you could just set hnf to be 'number' or 'id' to show hbond id on figure
    -genscript, --genscript (optional)
            whether to generate scripts for calculating distance and angle of hbonds
    )
    -cda, --calc_distance_angle (optional)
            whether to calculate distance and angle of hbonds from distance xvg file and angle xvg file
    -distancefile, --distancefile (optional)
            distance file of hbonds for input
    -anglefile, --anglefile (optional)
            angle file of hbonds for input
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xvg
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
    -x, --xlabel (optional)
            specify the xlabel of figures
    -so, --set_operation (optional)
            use AND or OR to perform set operation on hbonds, eg. `AND1,2,5-6`  
            or `OR0,2-4,7`. The AND set or OR set of hbonds whose ids were after
            key words will be calculated and shown in occupancy figure.  
""",
            "mol_map": """
mol_map:  map the coordinates of coor_file to atoms in name_file. The molecules
          in two pdb file need to share exactly same conformation.
          **YOU ARE NOT SUPPOSED TO USE THIS COMMAND!**

:examples:
    dit mol_map -n good.gro -c coor.pdb -o output.pdb

:parameters:
    -n, --name
            the pdb file which contains infos with coordinates you don't want
    -c, --coor
            the pdb file with coordinates and the conformation of molecule is 
            the same with the molecule in -n
    -o, --output
            the pdb file to write output
""",
            "dccm_ascii": """
dccm_ascii: generate dynamic cross correlation matrix (DCCM) from dat file 
            generated by `gmx covar`.

:examples:
    dit dccm_ascii -f covapic.dat 
    dit dccm_ascii -f covapic.dat -m gaussian -o res.png -noshow
    dit dccm_ascii -f covapic.dat -x Res -y Res -t DCCM

:parameters:
    -f, --input 
            the dat file generated by `gmx covar` which stored the covarience
            matrix information
    -o, --output (optional)
            specify the output figure name
    -m, --mode (optional)
            specify the figure style, "origin", "gaussian", "bio3d"(default)
    -ns, --noshow (optional)
            whether not to show figure
    -x, --xlabel (optional)
            specify the xlabel of figure
    -y, --ylabel (optional)
            specify the ylabel of figure
    -t, --title (optional)
            specify the title of figure
""",
            "show_style": """
show_style: generate matplotlib style template. 

:examples:
    dit show_style
""",
            "dssp": """
dssp: read in one xpm of secondary structure file, output the DSSP plot, 
      residue occupancy vs time figure, time occupancy vs residue figure. 

:examples:
    dit dssp -f dssp.xpm 
    dit dssp -f dssp.xpm -xs 0.001 -x Time(ns)

:parameters:
    -f, --input 
            input xpm of secondary structure
    -x, --xlabel (optional)
            specify the xlabel of xpm
    -y, --ylabel (optional)
            specify the ylabel of xpm
    -t, --title (optional)
            specify the title of xpm
    -xs, --xshrink (optional)
            specify a factor for multiplication of x-axis. default == 1.0
            For instance, if "-xs 0.001" is specified, all x-axis value of xvg
            will multiply this value. x-axis 1000 will be shown as 1. 
            Useful for converting the unit (ps) of time into (ns). Don't forget
            to change xlabel too after specifing -xs.
""",
        }

    def print_help_infos(self, method: str) -> None:
        """print help messages"""

        if method in self.help_infos.keys():
            print(self.help_infos[method])
        else:
            logging.error(
                "unknown command {}, type `dit help` for more infos.".format(method)
            )
            sys.exit()


def print_help_msg(method: str) -> None:
    """print help messages"""

    help_class = HELP()
    help_class.print_help_infos(method)


def get_welcome_msg() -> str:
    """return welcome messages"""

    welcome_info = """
 *******           **                  **********               **        
/**////**         /**          **   **/////**///               /**        
/**    /** **   **/** **    **//** **     /**  ******   ****** /**  ******
/**    /**/**  /**/**/**   /** //***      /** **////** **////**/** **//// 
/**    /**/**  /**/**//** /**   /**       /**/**   /**/**   /**/**//***** 
/**    ** /**  /**/** //****    **        /**/**   /**/**   /**/** /////**
/*******  //******/**  //**    **         /**//****** //****** *** ****** 
///////    ////// //    //    //          //  //////   ////// /// //////  

DuIvyTools is a simple analysis and visualization tool for GROMACS result files
written by CharlesHahn (https://github.com/CharlesHahn/DuIvyTools). 

DuIvyTools provides about 30 commands for visualization and processing of GMX
result files like .xvg or .xpm. 

All commands are shown below:
    XVG:
        xvg_show, xvg_compare, xvg_ave, xvg_mvave, xvg2csv, xvg_rama
        xvg_show_distribution, xvg_show_stack, xvg_show_scatter
        xvg_energy_compute, xvg_combine, xvg_ave_bar, xvg_box
    XPM:
        xpm_show, xpm2csv, xpm2dat, xpm2gpl
    NDX:
        ndx_show, ndx_rm_dup, ndx_rm, ndx_preserve
        ndx_add, ndx_combine, ndx_rename
    MDP:
        mdp_gen
    Others:
        find_center, pipi_dist_ang, hbond, mol_map, dccm_ascii, dssp
    Matplotlib Style:
        show_style

You can type `dit help <command>` or `dit <command> -h` for more help messages 
about each command, like: `dit help xvg_show` or `dit xvg_show -h`. 

And you can also modify the style of figures by adding (only) one mplstyle file
to your working directory. DIT will apply it to custom figures. You could 
modify mplstyle file based on the template generated by "dit show_style", or 
select one from style folder of DuIvyTools github repo 
(https://github.com/CharlesHahn/DuIvyTools). 

Have a good day !
"""
    return welcome_info


def help_call_functions(arguments: list = None) -> None:
    """call functions according to arguments"""

    if arguments == None:
        arguments = [argv for argv in sys.argv]
    description = get_welcome_msg()

    if len(arguments) < 2:
        DIT_infos = get_welcome_msg()
        print(DIT_infos)
        sys.exit()
    elif len(arguments) == 2:
        if arguments[1] in ["help", "-h", "--help"]:
            print(description)
        else:
            logging.error(
                "unknown method {}, type `dit help` for more infos.".format(
                    arguments[1]
                )
            )
            sys.exit()
    else:
        if arguments[1] != "help":
            logging.error("unknown command {}".format(arguments[1]))
            sys.exit()
        else:
            methods = arguments[2:]
            for method in methods:
                print_help_msg(method)


def main():
    help_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
FindCenter module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to provide some useful funcitons.

This module contains:
    - find_center function

This file is provided to you by GPLv3 license."""


import os
import sys
import argparse
import logging

base = os.path.dirname(os.path.realpath(__file__))
if base not in sys.path:
    sys.path.insert(0, base)

from NDX import NDX


def get_ndx(ndx_file: str = ""):
    ndx = NDX(ndx_file)
    ndx.show_ndx()
    select = input("Select one group: ")
    if select.isnumeric() and int(select) < ndx.group_number:
        index = ndx.group_index_list[int(select)]
        logging.info(
            "You selected group {} which contains {} atoms".format(
                ndx.group_name_list[int(select)], len(index)
            )
        )
    elif select in ndx.group_name_list:
        index = ndx.group_index_list[ndx.group_name_list.index(select)]
        logging.info(
            "You selected group {} which contains {} atoms".format(select, len(index))
        )
    else:
        logging.error("Wrong selection, type Int number or name of group")
        sys.exit()
    return index


def find_center(
    gro_file: str = "", ndx_file: str = None, AllAtoms: bool = None
) -> None:
    """
    find_center: a function to figure out one atom coordinates which
    is the closest one to group center, useful for adjusting PBC
    This function will read one gro file and return the center of coordinates
    and the atom number which is close to it.

    :parameters:
        gro_file: the gro file which saved coordinates

    :return:
        None
    """

    if not os.path.exists(gro_file):
        logging.error("no {} in current dirrectory. ".format(gro_file))
        sys.exit()
    if gro_file[-4:] != ".gro":
        logging.error("find_center only accept gro file with suffix .gro")
        sys.exit()
    if ndx_file != None and (not os.path.exists(ndx_file)):
        logging.error("no {} in current dirrectory. ".format(ndx_file))
        sys.exit()

    with open(gro_file, "r") as fo:
        lines = fo.readlines()
    atom_count = int(lines[1].strip())
    logging.info("{:d} atoms in {}".format(atom_count, gro_file))
    atom_lines = lines[2 : atom_count + 2]

    if ndx_file != None:
        index = get_ndx(ndx_file)
    else:
        index = [i for i in range(1, atom_count + 1)]

    atom_count = len(index)
    ## calculate the center point
    center_x, center_y, center_z = 0, 0, 0
    for i in index:
        line = atom_lines[i - 1]
        center_x += float(line[20:28])
        center_y += float(line[28:36])
        center_z += float(line[36:44])
    center_x = center_x / atom_count
    center_y = center_y / atom_count
    center_z = center_z / atom_count
    logging.info(
        "the center point is ({:.3f}, {:.3f}, {:.3f})".format(
            center_x, center_y, center_z
        )
    )

    ## find the closed atom
    atom_info = ""
    dist = 5  # nm, find closed atom in sphere of 5nm
    for i, line in enumerate(atom_lines):
        if (not AllAtoms) and (i + 1 not in index):
            continue
        x = float(line[20:28])
        y = float(line[28:36])
        z = float(line[36:44])
        atom_center_dist = (
            (x - center_x) ** 2 + (y - center_y) ** 2 + (z - center_z) ** 2
        ) ** 0.5
        if atom_center_dist < dist:
            dist = atom_center_dist
            atom_info = line[:44]
    if atom_info == "":
        logging.info("no atom detected in 5 nm sphere of center point. ")
        sys.exit()
    else:
        logging.info("distance from nearest atom to center: {:.3f} nm".format(dist))
        print("--------------------------------------------")
        print("ResID Name Atom  Num       X       Y       Z")
        print("--------------------------------------------")
        print(atom_info)
        print("--------------------------------------------")


def find_center_call_functions(arguments: list = []):
    """call functions according to arguments"""

    if len(arguments) == 0:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(description="Find the center of a group of atoms")
    parser.add_argument("-f", "--input", help="gro file for input")
    parser.add_argument("-n", "--index", help="index file for input")
    parser.add_argument(
        "-aa",
        "--AllAtoms",
        action="store_true",
        help="if to find center in all atoms of gro file",
    )

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        sys.exit()

    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        sys.exit()

    if len(arguments) == 2:
        logging.error("no parameters, type 'dit <command> -h' for more infos.")
        sys.exit()
    args = parser.parse_args(arguments[2:])
    if method == "find_center":
        print(args.input, args.index, args.AllAtoms)
        find_center(args.input, args.index, args.AllAtoms)
    else:
        logging.error("unknown method {}".format(method))
        sys.exit()

    logging.info("good day !")


def main():
    find_center_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
Visualizer_gnuplot module is part of DuIvyTools providing visualization tools based on gnuplot engine.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import time
import subprocess
from typing import List

import numpy as np

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log

## TODO maybe re-construction is needed for this module
## Gnuplot should support the functions like: plt.scatter() plt.lines(), plt.imshow()


class Gnuplot(log):
    """Gnuplot class for plotting with gnuplot"""

    def __init__(self) -> None:
        self.style: str = ""
        self.ntics: int = 8

        self.outfig: str = None
        self.title: str = None
        self.xlabel: str = None
        self.ylabel: str = None
        self.zlabel: str = None
        self.xmin: float = None
        self.xmax: float = None
        self.ymin: float = None
        self.ymax: float = None
        self.zmin: float = None
        self.zmax: float = None
        self.x_precision: int = None
        self.y_precision: int = None
        self.z_precision: int = None
        self.legends: List[str] = None
        self.xdata: List[List[float]] = None
        self.ydata: List[List[float]] = None
        self.data: List[List[float]] = None
        self.highs: List[List[float]] = None
        self.lows: List[List[float]] = None
        self.color_list: List[List[float]] = None
        self.colorbar_location: str = None
        self.legend_location: str = "inside"
        self.alpha = None
        self.colormap: str = None
        self.plot_type: str = "line"
        self.mode: str = None
        self.xtitles: List[str] = None
        self.stds_list: List[List[str]] = None

    def use_style(self, file: str) -> None:
        """load gnuplot style file in to Gnuplot class"""
        with open(file, "r") as fo:
            content = fo.read()
        self.style = content

    def check_repeat_values(self, values) -> bool:
        """True for repeat values exists in values"""
        if len(list(set(values))) < len(values):
            return True  # for repeat values
        else:
            return False

    def set_xy_repeat_tick_precision(self, gpl: str) -> None:
        """if repeat values in xaxis or yaxis, change to index of X and Y and set tics"""
        if self.check_repeat_values(self.xdata):
            self.info(
                "repeated values detected in xaxis, use index and set ticks by DIT"
            )
            x_step = len(self.xdata) // self.ntics
            x_step = [x_step, 1][x_step == 0]
            xdata_index = [i for i in range(0, len(self.xdata), x_step)]
            gpl += """set xtics ("""
            gpl += ",".join(
                [f""""{self.xdata[i]:.{self.x_precision}f}" {i}""" for i in xdata_index]
            )
            gpl += """)\n"""
            self.xdata = [i for i in range(len(self.xdata))]
        if self.check_repeat_values(self.ydata):
            self.info(
                "repeated values detected in yaxis, use index and set ticks by DIT"
            )
            y_step = len(self.ydata) // self.ntics
            y_step = [y_step, 1][y_step == 0]
            ydata_index = [i for i in range(0, len(self.ydata), y_step)]
            gpl += """set ytics ("""
            gpl += ",".join(
                [f""""{self.ydata[i]:.{self.y_precision}f}" {i}""" for i in ydata_index]
            )
            gpl += """)\n"""
            self.ydata = [i for i in range(len(self.ydata))]

        return gpl

    def dump2str(self) -> str:
        """dump gnuplot attributes to gnuplot input scripts string

        Returns:
            str: result string for gnuplto input
        """
        gpl: str = ""
        gpl += self.style + "\n"
        if self.outfig != None:
            gpl += f"""set output "{self.outfig}"\n"""
        if self.title != None:
            gpl += f"""set title "{self.title}"\n"""
        if self.xlabel != None:
            gpl += f"""set xlabel "{self.xlabel}"\n"""
        if self.ylabel != None:
            gpl += f"""set ylabel "{self.ylabel}"\n"""
        if self.zlabel != None:
            gpl += f"""set zlabel "{self.zlabel}"\n"""
        if self.legend_location == "inside":
            gpl += f"""set key inside\n"""
        elif self.legend_location == "outside":
            gpl += f"""set key outside reverse Left\n"""

        if self.xmin == None:
            self.xmin = ""
        if self.xmax == None:
            self.xmax = ""
        if self.ymin == None:
            self.ymin = ""
        if self.ymax == None:
            self.ymax = ""

        if self.x_precision != None:
            gpl += f"""set xtics format "%.{self.x_precision}f" \n"""
        if self.y_precision != None:
            gpl += f"""set ytics format "%.{self.y_precision}f" \n"""
        if self.z_precision != None:
            gpl += f"""set ztics format "%.{self.z_precision}f" \n"""

        if self.plot_type == "line":
            gpl = self.line_plot(gpl)
        elif self.plot_type == "scatter":
            gpl = self.scatter_plot(gpl)
        elif self.plot_type == "stack":
            gpl = self.stack_plot(gpl)
        elif self.plot_type == "violin":
            gpl = self.violin_plot(gpl)
        elif self.plot_type == "bar":
            gpl = self.bar_plot(gpl)
        elif self.plot_type == "imshow":
            gpl = self.imshow(gpl)
        elif self.plot_type == "3d":
            gpl = self.threeDimension(gpl)
        elif self.plot_type == "contour":
            gpl = self.contour(gpl)

        return gpl

    def threeDimension(self, gpl: str) -> str:
        """dump data of 3d plot to string"""
        if self.zlabel != None:
            gpl += f"""set cblabel "{self.zlabel}"\n"""
        if self.z_precision != None:
            gpl += f"""set cbtics format "%.{self.z_precision}f"\n"""
        gpl += "set pm3d implicit at s\n"
        gpl += "set colorbox user\n"
        gpl += "set contour base\n"
        gpl += "set cntrparam levels 10\n"
        # gpl += "set ylabel norotate offset -1,0\n"
        gpl += "set colorbox vertical origin 0.9, 0.2 size 0.03, 0.6 front noinvert noborder\n"

        gpl = self.set_xy_repeat_tick_precision(gpl)

        gpl += f"\n$matrix << EOD\n"
        for y, y_value in enumerate(self.ydata):
            for x, x_value in enumerate(self.xdata):
                gpl += f"""{x_value} {y_value} {self.data[y][x]}\n"""
            gpl += "\n"
        gpl += "EOD\n\n"
        gpl += f"splot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
        gpl += f"""$matrix using 1:2:3 with pm3d notitle, \\\n"""
        gpl += f"""$matrix using 1:2:3 with lines nosurface notitle """
        gpl += "\n"

        return gpl

    def contour(self, gpl: str) -> str:
        """dump data of contour plot to string"""
        if self.zlabel != None:
            gpl += f"""set cblabel "{self.zlabel}"\n"""
        if self.z_precision != None:
            gpl += f"""set cbtics format "%.{self.z_precision}f"\n"""
        gpl += "set pm3d implicit at s\n"
        gpl += "set colorbox user\n"
        gpl += "set view map\n"
        gpl += "set ylabel norotate offset -1,0\n"
        gpl += "set contour base\n"
        gpl += "set colorbox vertical origin screen 0.9, 0.2 size screen 0.03, 0.6 front  noinvert noborder\n"

        gpl = self.set_xy_repeat_tick_precision(gpl)

        gpl += f"\n$matrix << EOD\n"
        for y, y_value in enumerate(self.ydata):
            for x, x_value in enumerate(self.xdata):
                gpl += f"""{x_value} {y_value} {self.data[y][x]}\n"""
            gpl += "\n"
        gpl += "EOD\n\n"
        gpl += f"splot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
        gpl += f"""$matrix using 1:2:3 with lines nosurface notitle """
        gpl += "\n"

        return gpl

    def imshow(self, gpl: str) -> str:
        """dump data of image plot to string"""
        if self.xpm_type != "Continuous":
            gpl += "set key out reverse Left spacing 2 samplen 1/2\n"
            gpl += "unset colorbox\n"
            pal_line = "set pal defined("
            for index, color in enumerate(self.color_list):
                pal_line += f"""{index} "{color}","""
            pal_line = pal_line.strip(",") + ")"
            gpl += pal_line + "\n\n"

            gpl = self.set_xy_repeat_tick_precision(gpl)

            gpl += f"\n$matrix << EOD\n"
            for y, y_value in enumerate(self.ydata):
                for x, x_value in enumerate(self.xdata):
                    gpl += f"""{x_value} {y_value} {self.data[y][x]}\n"""
                gpl += "\n"
            gpl += "EOD\n\n"
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            gpl += f"""$matrix using 1:2:3 with image notitle, \\\n"""
            for id, leg in enumerate(self.legends):
                gpl += f"""{np.floor(np.min(self.ydata))-1} w p ps 4 pt 5 lc rgb "{self.color_list[id]}" title "{leg}", \\\n"""
            gpl = gpl.strip("\n").strip("\\").strip().strip(",")

        else:
            if self.zlabel != None:
                gpl += f"""set cblabel "{self.zlabel}"\n"""
            if self.z_precision != None:
                gpl += f"""set cbtics format "%.{self.z_precision}f"\n"""

            gpl = self.set_xy_repeat_tick_precision(gpl)

            gpl += f"\n$matrix << EOD\n"
            for y, y_value in enumerate(self.ydata):
                for x, x_value in enumerate(self.xdata):
                    gpl += f"""{x_value} {y_value} {self.data[y][x]}\n"""
                gpl += "\n"
            gpl += "EOD\n\n"
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            gpl += f"""$matrix using 1:2:3 with image notitle """
            gpl += "\n"

        return gpl

    def stack_plot(self, gpl: str) -> str:
        """dump data of stack line plot to string"""
        gpl += f"""set style fill transparent solid {self.alpha} noborder\n"""
        for c in range(len(self.data)):
            gpl += f"\n$data{c} << EOD\n"
            for r in range(len(self.xdata[c])):
                gpl += f"""{self.xdata[c][r]} {self.data[c][r]} {self.highs[c][r]} {self.lows[c][r]}\n"""
            gpl += "EOD\n\n"
        gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
        for c in range(len(self.data)):
            gpl += f"""$data{c} using 1:3:4 with filledcurves linestyle {c+1} title "{self.legends[c]}", \\\n"""
        gpl += "\n"

        return gpl

    def violin_plot(self, gpl: str) -> str:
        """dump data of violin plot to string"""
        if self.data and self.legends:
            for c in range(len(self.data)):
                gpl += f"\n$data{c} << EOD\n"
                for r in range(len(self.data[c])):
                    gpl += f"""{self.data[c][r]} {self.color_list[c][r]}\n"""
                gpl += "EOD\n\n"

            gpl += """set errorbars front 1.000000  lt black linewidth 1.000 dashtype solid
set boxwidth 0.075 absolute
set style fill solid 1.00 border lt -1
set style data filledcurves below 
set colorbox vertical origin screen 0.9, 0.2 size screen 0.05, 0.6 front  noinvert bdefault\n"""
            gpl += """set xtics ("""
            gpl += ",".join(
                [f""""{leg}" {i+1}""" for i, leg in enumerate(self.legends)]
            )
            gpl += """)\n"""
            for c in range(len(self.data)):
                gpl += f"""set table $kdedata{c}\n"""
                gpl += f"""plot $data{c} using 1:(1) smooth kdensity with filledcurves above y lt {c+1} title "{self.legends[c]}"\n"""
            gpl += "unset table\n"
            gpl += "unset key\n"
            for c in range(len(self.data)):
                gpl += (
                    f"""stats $kdedata{c} using 2 nooutput name "kdedata{c}_stats" \n"""
                )
            ## scatter plot
            loc = 1.0
            if self.mode != "withoutScatter":
                loc = 0.75
                if self.zlabel != None:
                    gpl += f"""set cblabel "{self.zlabel}"\n"""
                if self.z_precision != None:
                    gpl += f"""set cbtics format "%.{self.z_precision}f"\n"""
                gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
                for c in range(len(self.data)):
                    gpl += f"""$data{c} u ({c+1.25}+0.04*invnorm(rand(0))):1:2 title "{self.legends[c]}" with points palette, \\\n"""

            if self.mode == "withoutScatter":
                gpl += f"\nplot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            for c in range(len(self.data)):
                gpl += f"""$kdedata{c} using ({c+loc}+$2/(4.2*kdedata{c}_stats_max)):1 with filledcurve x={c+loc} lt {c+1},\\\n"""
                gpl += f"""$kdedata{c} using ({c+loc}-$2/(4.2*kdedata{c}_stats_max)):1 with filledcurve x={c+loc} lt {c+1},\\\n"""
                gpl += f"""$data{c} u ({c+loc}):1 with boxplot fc "white" lw 1 ,\\\n"""
            gpl += "\n"

        return gpl

    def scatter_plot(self, gpl: str) -> str:
        """dump data of scatter plot to string"""
        # TODO colorbar_location
        if self.zlabel != None:
            gpl += f"""set cblabel "{self.zlabel}"\n"""
        if self.z_precision != None:
            gpl += f"""set cbtics format "%.{self.z_precision}f"\n"""
        if self.alpha == None:
            self.alpha = 1.0
        gpl += f"set style fill transparent solid {self.alpha} noborder\n"
        if self.data and self.legends:
            for c in range(len(self.data)):
                colors = self.color_list[c]
                if colors != None:
                    gpl += f"\n$data{c} << EOD\n"
                    for r in range(len(self.xdata[c])):
                        gpl += f"""{self.xdata[c][r]} {self.data[c][r]} {colors[r]}\n"""
                    gpl += "EOD\n\n"
                else:
                    gpl += f"\n$data{c} << EOD\n"
                    for r in range(len(self.xdata[c])):
                        gpl += f"""{self.xdata[c][r]} {self.data[c][r]} \n"""
                    gpl += "EOD\n\n"
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            for c in range(len(self.data)):
                colors = self.color_list[c]
                if colors != None:
                    gpl += f"""$data{c} u 1:2:3 title "{self.legends[c]}" with points palette, \\\n"""
                else:
                    gpl += f"""$data{c} u 1:2 title "{self.legends[c]}" with points, \\\n"""
            gpl += "\n"

        return gpl

    def line_plot(self, gpl: str) -> str:
        """dump data of line plot to string"""
        if self.data and self.legends and len(self.highs) == 0 and len(self.lows) == 0:
            for c in range(len(self.data)):
                gpl += f"\n$data{c} << EOD\n"
                for r in range(len(self.xdata[c])):
                    gpl += f"""{self.xdata[c][r]} {self.data[c][r]}\n"""
                gpl += "EOD\n\n"
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            for c in range(len(self.data)):
                gpl += f"""$data{c} u 1:2 title "{self.legends[c]}" with lines linestyle {c+1}, \\\n"""
            gpl += "\n"

        if self.data and self.legends and len(self.highs) != 0 and len(self.lows) != 0:
            gpl += f"""set style fill transparent solid {self.alpha} noborder\n"""
            for c in range(len(self.data)):
                gpl += f"\n$data{c} << EOD\n"
                for r in range(len(self.xdata[c])):
                    gpl += f"""{self.xdata[c][r]} {self.data[c][r]} {self.highs[c][r]} {self.lows[c][r]}\n"""
                gpl += "EOD\n\n"
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            for c in range(len(self.data)):
                gpl += f"""$data{c} using 1:3:4 with filledcurves notitle linestyle {c+1}, $data{c} u 1:2 title "{self.legends[c]}" with lines linestyle {c+1}, \\\n"""
            gpl += "\n"

        return gpl

    def bar_plot(self, gpl: str) -> str:
        """dump data of bar plot to string"""
        if self.data and self.legends:
            for c in range(len(self.data)):
                gpl += f"\n$data{c} << EOD\n"
                xdata = [x for x in range(len(self.data[c]))]
                for x, d, s in zip(xdata, self.data[c], self.stds_list[c]):
                    gpl += f"""{x} {d} {s}\n"""
                gpl += "EOD\n\n"
            gpl += """set xtics ("""
            gpl += ",".join([f""""{xt}" {i}""" for i, xt in enumerate(self.xtitles)])
            gpl += """)\n"""
            gpl += """set style histogram errorbars lw 2\n"""
            gpl += """set style fill solid border -1\n"""
            gpl += f"plot [{self.xmin}:{self.xmax}][{self.ymin}:{self.ymax}] "
            for c in range(len(self.data)):
                gpl += f"""$data{c} u 2:3 title "{self.legends[c]}" with histogram linestyle {c+1}, \\\n"""
            gpl += "\n"

        return gpl


class ParentGnuplot(log):
    """the parent class of gnuplot visualizer classes"""

    def __init__(self) -> None:
        time_info = time.strftime("%Y%m%d%H%M%S", time.localtime())
        self.outfig: str = f"DIT_gnuplot_output_{time_info}.png"
        self.gpl_file: str = f"DIT_gnuplot_script_{time_info}.gnu"
        self.gnuplot = Gnuplot()
        self.load_style()

    def load_style(self):
        """load gnuplot style file"""
        style_files = [file for file in os.listdir() if file[-8:] == ".gpstyle"]
        if len(style_files) == 1:
            self.gnuplot.use_style(style_files[0])
            self.info(f"using gnuplot style from {style_files[0]}")
        elif len(style_files) > 1:
            self.gnuplot.use_style(style_files[0])
            self.info(
                f"more than one gnuplot style files detected, using the {style_files[0]}"
            )
        else:
            data_file_path = os.path.realpath(
                os.path.join(os.getcwd(), os.path.dirname(__file__), "../")
            )
            mplstyle = os.path.join(
                data_file_path, os.path.join("data", "gnuplotstyle", "DIT.gpstyle")
            )
            self.gnuplot.use_style(mplstyle)
            self.info(
                "using default gnuplot style sheet, to inspect its content, use 'dit show_style -eg gnuplot'"
            )

    def dump(self) -> None:
        """dump gnuplot input scripts into a input file"""
        gpl = self.gnuplot.dump2str()
        with open(self.gpl_file, "w") as fo:
            fo.write(gpl)
        self.info(f"temporarily dump gnuplot scripts to {self.gpl_file}")

    def clean(self) -> None:
        """remove the gnuplot input script file"""
        os.remove(self.gpl_file)
        self.info(f"removed gnuplot scripts {self.gpl_file}")

    def run_file(self) -> None:
        """run the gnuplot script to get figure"""
        inCmd = f"""echo load "{self.gpl_file}" | gnuplot"""
        p = subprocess.Popen(
            inCmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        self.info(f"running gnuplot at pid -> {p.pid}")
        output, error = p.communicate()
        p.wait()
        status = ("Fail", "Success")[p.returncode == 0]
        output, error = output.decode(), error.decode()
        self.info(f"gnuplot status -> {status}")
        if output:
            self.info(f"gnuplot output -> {output}")
        if error:
            self.error(f"gnuplot error -> {error}")

    def run(self) -> None:
        """run the gnuplot to get figure"""
        gpl = self.gnuplot.dump2str()
        cmd = gpl.encode()
        p = subprocess.Popen(["gnuplot"], stdin=subprocess.PIPE)
        p.stdin.write(cmd)
        p.stdin.close()
        self.info(f"running gnuplot at pid -> {p.pid}")
        output, error = p.communicate()
        p.wait()
        status = ("Fail", "Success")[p.returncode == 0]
        self.info(f"gnuplot status -> {status}")
        if output:
            self.info(f"gnuplot output -> {output.decode()}")
        if error:
            self.error(f"gnuplot error -> {error.decode()}")

    def final(self, outfig: str, noshow: bool) -> None:
        """deal with final process of plotting by gnuplot

        Args:
            outfig (str): the user specified output figure name
            noshow (bool): True for not delete the gnuplot scripts file
        """
        if outfig != None:
            if os.path.exists(outfig):
                time_info = time.strftime("%Y%m%d%H%M%S", time.localtime())
                new_outfig = f'{".".join(outfig.split(".")[:-1])}_{time_info}.{outfig.split(".")[-1]}'
                self.warn(
                    f"{outfig} is already in current directory, save to {new_outfig} for instead."
                )
                outfig = new_outfig
            self.outfig = outfig
        self.gnuplot.outfig = self.outfig
        if not noshow:
            self.run()
            # self.dump()
            # self.run_file()
            # self.clean()
        else:
            self.dump()


class LineGnuplot(ParentGnuplot):
    """A Gnuplot line plot class for line plots

    Args:
        ParentGnuplot (object): Gnuplot parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        highs :List[List[float]]
        lows :List[List[float]]
        alpha :float
        legend_location :str # {inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        self.gnuplot.title = kwargs["title"]
        self.gnuplot.xlabel = kwargs["xlabel"]
        self.gnuplot.ylabel = kwargs["ylabel"]

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            self.gnuplot.xmin = kwargs["xmin"]
            self.gnuplot.xmax = kwargs["xmax"]
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            self.gnuplot.ymin = kwargs["ymin"]
            self.gnuplot.ymax = kwargs["ymax"]
        if kwargs["x_precision"] != None:
            self.gnuplot.x_precision = kwargs["x_precision"]
        if kwargs["y_precision"] != None:
            self.gnuplot.y_precision = kwargs["y_precision"]

        if len(kwargs["legends"]) != len(kwargs["data_list"]):
            self.error(
                f"""unable to pair {len(kwargs["legends"])} legends to {len(kwargs["data_list"])} column data."""
            )
        self.gnuplot.xdata = kwargs["xdata_list"]
        self.gnuplot.data = kwargs["data_list"]
        self.gnuplot.legends = kwargs["legends"]
        self.gnuplot.highs = kwargs["highs"]
        self.gnuplot.lows = kwargs["lows"]
        self.gnuplot.alpha = kwargs["alpha"]
        self.gnuplot.legend_location = kwargs["legend_location"]


class StackGnuplot(LineGnuplot):
    """A Gnuplot stack line plot class for stack line plots

    Args:
        ParentGnuplot (object): Gnuplot parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        highs :List[List[float]]
        lows :List[List[float]]
        alpha :float
        legend_location :str # {inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.gnuplot.plot_type = "stack"


class ScatterGnuplot(ParentGnuplot):
    """A gnuplot scatter plot class for scatter plots

    Args:
        ParentGnuplot (object): gnuplot parent class

    Parameters:
        data_list :List[List[float]]
        xdata_list :List[List[float]]
        color_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        zmin :float
        zmax :float
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        cmap :str
        colorbar_location:str
        legend_location :str # {inside, outside}
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        self.gnuplot.title = kwargs["title"]
        self.gnuplot.xlabel = kwargs["xlabel"]
        self.gnuplot.ylabel = kwargs["ylabel"]

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            self.gnuplot.xmin = kwargs["xmin"]
            self.gnuplot.xmax = kwargs["xmax"]
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            self.gnuplot.ymin = kwargs["ymin"]
            self.gnuplot.ymax = kwargs["ymax"]
        self.gnuplot.x_precision = kwargs["x_precision"]
        self.gnuplot.y_precision = kwargs["y_precision"]
        self.gnuplot.z_precision = kwargs["z_precision"]

        if len(kwargs["legends"]) != len(kwargs["data_list"]):
            self.error(
                f"""unable to pair {len(kwargs["legends"])} legends to {len(kwargs["data_list"])} column data."""
            )
        self.gnuplot.xdata = kwargs["xdata_list"]
        self.gnuplot.data = kwargs["data_list"]
        self.gnuplot.legends = kwargs["legends"]
        self.gnuplot.color_list = kwargs["color_list"]
        self.gnuplot.zlabel = kwargs["zlabel"]
        self.gnuplot.plot_type = "scatter"
        self.gnuplot.alpha = kwargs["alpha"]
        self.gnuplot.legend_location = kwargs["legend_location"]
        # self.gnuplot.colormap = kwargs["cmap"]
        if kwargs["cmap"]:
            self.warn(
                "DIT is unable to set colormap for gnuplot now. The https://github.com/Gnuplotting/gnuplot-palettes may help you"
            )
        # self.gnuplot.colorbar_location = kwargs["colorbar_location"]
        if kwargs["colorbar_location"]:
            self.warn("DIT is unable to set colorbar location for gnuplot now.")


class BarGnuplot(ParentGnuplot):
    """A Gnuplot bar plot class for bar plots

    Args:
        ParentGnuplot (object): gnuplot parent class

    Parameters:
        data_list :List[List[float]]
        stds_list :List[List[float]]
        xtitles :List[str]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        title :str
        x_precision :int
        y_precision :int
        legend_location :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        self.gnuplot.title = kwargs["title"]
        self.gnuplot.xlabel = kwargs["xlabel"]
        self.gnuplot.ylabel = kwargs["ylabel"]

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            self.gnuplot.xmin = kwargs["xmin"]
            self.gnuplot.xmax = kwargs["xmax"]
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            self.gnuplot.ymin = kwargs["ymin"]
            self.gnuplot.ymax = kwargs["ymax"]
        if kwargs["x_precision"] != None:
            self.gnuplot.x_precision = kwargs["x_precision"]
        if kwargs["y_precision"] != None:
            self.gnuplot.y_precision = kwargs["y_precision"]

        if len(kwargs["legends"]) != len(kwargs["data_list"]):
            self.error(
                f"""unable to pair {len(kwargs["legends"])} legends to {len(kwargs["data_list"])} column data."""
            )
        self.gnuplot.data = kwargs["data_list"]
        self.gnuplot.legends = kwargs["legends"]
        self.gnuplot.legend_location = kwargs["legend_location"]
        self.gnuplot.xtitles = kwargs["xtitles"]
        self.gnuplot.stds_list = kwargs["stds_list"]
        self.gnuplot.plot_type = "bar"


class BoxGnuplot(ParentGnuplot):
    """A gnuplot box plot class for box plots

    Args:
        ParentGnuplot (object): gnuplot parent class

    Parameters:
        data_list :List[List[float]]
        color_list :List[List[float]]
        legends :List[str]
        xmin :float
        xmax :flaot
        ymin :float
        ymax :float
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        x_precision :int
        y_precision :int
        z_precision :int
        alpha :float
        cmap :str
        colorbar_location:str
        mode :str
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()

        self.gnuplot.title = kwargs["title"]
        self.gnuplot.xlabel = kwargs["xlabel"]
        self.gnuplot.ylabel = kwargs["ylabel"]

        if kwargs["xmin"] != None or kwargs["xmax"] != None:
            self.gnuplot.xmin = kwargs["xmin"]
            self.gnuplot.xmax = kwargs["xmax"]
        if kwargs["ymin"] != None or kwargs["ymax"] != None:
            self.gnuplot.ymin = kwargs["ymin"]
            self.gnuplot.ymax = kwargs["ymax"]
        self.gnuplot.x_precision = kwargs["x_precision"]
        self.gnuplot.y_precision = kwargs["y_precision"]
        self.gnuplot.z_precision = kwargs["z_precision"]

        if len(kwargs["legends"]) != len(kwargs["data_list"]):
            self.error(
                f"""unable to pair {len(kwargs["legends"])} legends to {len(kwargs["data_list"])} column data."""
            )

        self.gnuplot.data = kwargs["data_list"]
        self.gnuplot.legends = kwargs["legends"]
        self.gnuplot.color_list = kwargs["color_list"]
        self.gnuplot.zlabel = kwargs["zlabel"]
        self.gnuplot.plot_type = "violin"
        self.gnuplot.mode = kwargs["mode"]
        # self.gnuplot.colormap = kwargs["cmap"]
        if kwargs["cmap"]:
            self.warn(
                "DIT is unable to set colormap for gnuplot now. The https://github.com/Gnuplotting/gnuplot-palettes may help you"
            )
        # self.gnuplot.colorbar_location = kwargs["colorbar_location"]
        if kwargs["colorbar_location"]:
            self.warn("DIT is unable to set colorbar location for gnuplot now.")


class ImshowGnuplot(ParentGnuplot):
    """A gnuplot imshow plot class for heatmap

    Args:
        ParentGnuplot (object): matplotlib parent class

    Parameters:
        fig_type :str
        data_list :List[List[float]]
        xdata_list :List[float]
        ydata_list :List[float]
        legends :List[str]
        color_list :List[str]
        xlabel :str
        ylabel :str
        zlabel :str
        title :str
        xmin :float
        xmax :float
        ymin :float
        ymax :float
        x_precision :int
        y_precision :int
        z_precision :int
        colorbar_location :str
        cmap :str
    """

    def __init__(self, plot_type: str, **kwargs) -> None:
        super().__init__()

        self.gnuplot.title = kwargs["title"]
        self.gnuplot.xlabel = kwargs["xlabel"]
        self.gnuplot.ylabel = kwargs["ylabel"]
        self.gnuplot.zlabel = kwargs["zlabel"]

        if len(kwargs["xdata_list"]) <= 1 or len(kwargs["ydata_list"]) <= 1:
            self.error(
                f"""Gnuplot engine can not handle data with 1 dimension ({len(kwargs["xdata_list"])}*{len(kwargs["ydata_list"])})"""
            )

        if len(list(set(kwargs["xdata_list"]))) == len(kwargs["xdata_list"]):
            data = kwargs["xdata_list"]
            dot_len_x = (np.max(data) - np.min(data)) / (len(data) - 1)
            if self.gnuplot.xmin == None:
                self.gnuplot.xmin = np.min(kwargs["xdata_list"]) - 0.5 * dot_len_x
            if self.gnuplot.xmax == None:
                self.gnuplot.xmax = np.max(kwargs["xdata_list"]) + 0.5 * dot_len_x
        else:
            self.gnuplot.xmin = -0.5
            self.gnuplot.xmax = len(kwargs["xdata_list"]) - 0.5
        if len(list(set(kwargs["ydata_list"]))) == len(kwargs["ydata_list"]):
            data = kwargs["ydata_list"]
            dot_len_y = (np.max(data) - np.min(data)) / (len(data) - 1)
            if self.gnuplot.ymin == None:
                self.gnuplot.ymin = np.min(kwargs["ydata_list"]) - 0.5 * dot_len_y
            if self.gnuplot.ymax == None:
                self.gnuplot.ymax = np.max(kwargs["ydata_list"]) + 0.5 * dot_len_y
        else:
            self.gnuplot.ymin = -0.5
            self.gnuplot.ymax = len(kwargs["ydata_list"]) - 0.5

        self.gnuplot.x_precision = kwargs["x_precision"]
        self.gnuplot.y_precision = kwargs["y_precision"]
        self.gnuplot.z_precision = kwargs["z_precision"]

        self.gnuplot.data = kwargs["data_list"]
        self.gnuplot.xdata = kwargs["xdata_list"]
        self.gnuplot.ydata = kwargs["ydata_list"]
        self.gnuplot.legends = kwargs["legends"]
        self.gnuplot.color_list = kwargs["color_list"]
        self.gnuplot.plot_type = plot_type
        self.gnuplot.xpm_type = kwargs["fig_type"]

        # self.gnuplot.colormap = kwargs["cmap"]
        if kwargs["cmap"]:
            self.warn(
                "DIT is unable to set colormap for gnuplot now. The https://github.com/Gnuplotting/gnuplot-palettes may help you"
            )
        # self.gnuplot.colorbar_location = kwargs["colorbar_location"]
        if kwargs["colorbar_location"]:
            self.warn("DIT is unable to set colorbar location for gnuplot now.")



"""
DIT module is part of DuIvyTools. This module provides log parent class.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys
import inspect

base = os.path.dirname(os.path.realpath(__file__))
if base not in sys.path:
    sys.path.insert(0, base)

from Commands.otherCommands import *
from Commands.xpmCommands import *
from Commands.xvgCommands import *
from utils import Parameters, log


class DIT(log):
    """DIT class, show information and invoke commands"""

    def __init__(self) -> None:
        self.classes = dict(inspect.getmembers(sys.modules[__name__], inspect.isclass))
        self.cmds = [
            "xvg_show",
            "xvg_compare",
            "xvg_ave",
            "xvg_energy_compute",
            "xvg_combine",
            "xvg_show_distribution",
            "xvg_show_scatter",
            "xvg_show_stack",
            "xvg_box_compare",
            "xvg_ave_bar",
            "xvg_rama",
            "xpm_show",
            "xpm2csv",
            "xpm2dat",
            "xpm_diff",
            "xpm_merge",
            "mdp_gen",
            "show_style",
            "find_center",
            "dccm_ascii",
            "dssp",
            "ndx_add",
            "ndx_split",
            "ndx_show",
        ]
        self.cmds_infos = """
XVG:
    xvg_show              : easily show xvg file
    xvg_compare           : visualize xvg data
    xvg_ave               : calculate the averages of xvg data
    xvg_energy_compute    : calculate eneries between protein and ligand
    xvg_combine           : combine data of xvg files
    xvg_show_distribution : show distribution of xvg data
    xvg_show_scatter      : show xvg data by scatter plot
    xvg_show_stack        : show xvg data by stack area plot
    xvg_box_compare       : compare xvg data by violin and scatter plots
    xvg_ave_bar           : calculate and show the averages of parallelism 
    xvg_rama              : draw ramachandran plot from xvg data
XPM:
    xpm_show              : visualize xpm data
    xpm2csv               : convert xpm data into csv file in form (x, y, z)
    xpm2dat               : convert xpm data into dat file in form (N*N)
    xpm_diff              : calculate the difference of xpms
    xpm_merge             : merge two xpm by half and half
Others:
    mdp_gen               : generate mdp file templates
    show_style            : show figure control style files
    find_center           : find geometric center of one group of atoms
    dccm_ascii            : convert dccm from ascii data file to xpm
    dssp                  : generate xpm and xvg from ascii file of gmx2023
    ndx_add               : new a index group to ndx file
    ndx_split             : split one index group into several groups
    ndx_show              : show the groupnames of index file
"""
        self.welcome_info = (
            """
 *******           **                  **********               **        
/**////**         /**          **   **/////**///               /**        
/**    /** **   **/** **    **//** **     /**  ******   ****** /**  ******
/**    /**/**  /**/**/**   /** //***      /** **////** **////**/** **//// 
/**    /**/**  /**/**//** /**   /**       /**/**   /**/**   /**/**//***** 
/**    ** /**  /**/** //****    **        /**/**   /**/**   /**/** /////**
/*******  //******/**  //**    **         /**//****** //****** *** ****** 
///////    ////// //    //    //          //  //////   ////// /// //////  

DuIvyTools is a simple analysis and visualization tool for GROMACS result files written by 杜艾维 (https://github.com/CharlesHahn/DuIvyTools). 

DuIvyTools provides about 30 commands for visualization and processing of GMX result files like .xvg or .xpm. 

All commands are shown below: """
            + self.cmds_infos
            + """
You can type `dit <command> -h` for detailed help messages about each command, like: `dit xvg_show -h`. 

All possible parameters could be inspected by `dit -h` or `dit --help`.

Cite DuIvyTools by DOI at https://doi.org/10.5281/zenodo.6339993

Have a good day !
"""
        )

    def run(self) -> None:
        """show help infos and invoke commands"""
        ## help infos
        if len(sys.argv) == 1:
            print(self.welcome_info)
            sys.exit()
        if len(sys.argv) == 3 and sys.argv[2] in ["-h", "--help", "help"]:
            if sys.argv[1] in self.cmds:
                print(f"====== command: {sys.argv[1]} ======")
                print(self.classes[sys.argv[1]].__doc__)
            else:
                self.error(
                    f"Wrong specification of command `{sys.argv[1]}`, type `dit` to see all possible commands"
                )
            sys.exit()

        ## try to run DIT
        parm = Parameters()
        if parm.cmd not in self.cmds:
            self.error(
                f"{parm.cmd} is not available. DIT supports commands as below: \n"
                + self.cmds_infos
            )
        cmd = self.classes.get(parm.cmd, None)
        cmd = cmd(parm)
        cmd()

def main():
    dit = DIT()
    dit.run()
    dit.info("May you good day !")

if __name__ == "__main__":
    main()



"""
hbond module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

This module is designed to process hbond information.

This module contains:
    - hbond function

This file is provided to you by GPLv3 license."""


import os
import sys
import logging
import argparse
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from pathlib import Path

base = os.path.dirname(os.path.realpath(__file__))
if base not in sys.path:
    sys.path.insert(0, base)

from XPM import XPM
from XVG import XVG


def gen_distang_script(donor_ndxs, hydrogen_ndxs, acceptor_ndxs, select):
    """generate scripts to calculate distance and angles of hbond

    Args:
        donor_ndxs (list): donor atom indexs
        hydrogen_ndxs (list): hydrogen atom indexs
        acceptor_ndxs (list): acceptor atom indexs
        select (list): the selected hbond ids
    """

    ## calc distance of donor - acceptor
    for ndx in ["hbdist.ndx", "hbang.ndx"]:
        if ndx in os.listdir():
            logging.error("{} is already in current directory!".format(ndx))
            sys.exit()

    tail = -1
    with open("hbdist.ndx", "w") as fo:
        fo.write("[ hbdist ] \n")
        for i in select:
            if donor_ndxs[i] != tail:
                fo.write("{}  {} \n".format(donor_ndxs[i], acceptor_ndxs[i]))
                tail = acceptor_ndxs[i]
            else:
                fo.write("{}  {} \n".format(acceptor_ndxs[i], donor_ndxs[i]))
                tail = donor_ndxs[i]

    ## gmx distance -f xtc -s tpr -n hbdist.ndx -oall hbdistall.xvg -select '"group"'
    hbdist_script = "run_hbdist.sh"
    if hbdist_script in os.listdir():
        logging.error("{} is already in current directory!".format(hbdist_script))
        sys.exit()
    with open(hbdist_script, "w") as fo:
        fo.write(
            """gmx distance -f xxx.xtc -s xxx.tpr -n hbdist.ndx -oall hbdistall.xvg -select '"hbdist"'"""
        )

    ## calc angle : hydrogen - donor - acceptor
    with open("hbang.ndx", "w") as fo:
        fo.write("[ hbang ] \n")
        for i in select:
            fo.write(
                "{} {} {} \n".format(hydrogen_ndxs[i], donor_ndxs[i], acceptor_ndxs[i])
            )
    ## gmx angle -f xxx.xtc -n hbangndx.ndx -ov hbangall.xvg -all -od angdist.xvg
    hbang_script = "run_hbang.sh"
    if hbang_script in os.listdir():
        logging.error("{} is already in current directory!".format(hbang_script))
        sys.exit()
    with open(hbang_script, "w") as fo:
        fo.write("""gmx angle -f xxx.xtc -n hbang.ndx -ov hbangall.xvg -all""")

    logging.info(
        """
        Please notice: 
            1. if you applied -dt, -b or -e in your command when calculating `gmx hbond`, add same parameters to scripts when calculating distance or angle of hbonds.
            2. YOU MUST apply `-merge no` to `gmx hbond` to get correct atom ids which if critically IMPORTANT for calculating angles and distances of hbonds. 
        """
    )


def calculate_distance_angle(
    xpm_datamatrix: list,
    hbond_names: list,
    distfile: str,
    angfile: str,
    figout: str = None,
    noshow: bool = False,
    xshrink: float = 1,
    xlabel: str = None,
):
    """calculate average distance and angle, draw figures.

    Args:
        xpm_datamatrix (list): source data
        hbond_names (list): name of hbonds
        distfile (str): file name of distances
        angfile (str): file name of angles
    """
    if len(xpm_datamatrix) != len(hbond_names):
        logging.error(
            "wrong in length of datamatrix and hbond names while calculating distances and angles"
        )
        sys.exit()
    if xshrink == None:
        xshrink = 1.0
    print(xshrink)
    dist_xvg = XVG(distfile)
    ang_xvg = XVG(angfile)
    if dist_xvg.xvg_column_num != len(xpm_datamatrix) + 1:
        logging.error("wrong in column number of distance xvg file, check select!")
        sys.exit()
    if ang_xvg.xvg_column_num != len(xpm_datamatrix) + 2:
        logging.error("wrong in column number of distance xvg file, check select!")
        sys.exit()
    if dist_xvg.xvg_row_num != ang_xvg.xvg_row_num != len(xpm_datamatrix[0]):
        logging.error("wrong in length of row, check time selection clearfully!")
        sys.exit()

    ## calc average distance and angle
    dist_ave_std, ang_ave_std = [], []
    for i, dot_line in enumerate(xpm_datamatrix):
        dot_line = np.array(dot_line)
        dist_line = dist_xvg.xvg_columns[i + 1]
        ang_line = ang_xvg.xvg_columns[i + 2]
        dist_present = np.array(dist_line)[dot_line == 1].astype(float)
        ang_present = np.array(ang_line)[dot_line == 1].astype(float)
        dist_ave = np.average(dist_present)
        dist_std = np.std(dist_present, ddof=1)
        ang_ave = np.average(ang_present)
        ang_std = np.std(ang_present, ddof=1)
        dist_ave_std.append((dist_ave, dist_std))
        ang_ave_std.append((ang_ave, ang_std))

    ## draw distance and angle figure
    plt.clf()
    dist_time = np.array(dist_xvg.xvg_columns[0]).astype(float)
    dist_time = dist_time * xshrink
    for i, hbond_name in enumerate(hbond_names):
        dist_line = np.array(dist_xvg.xvg_columns[i + 1]).astype(float)
        plt.plot(dist_time, dist_line, label=hbond_name)
    # plt.plot(dist_time, [0.35 for d in dist_time], "grey")
    if xlabel != None:
        plt.xlabel(xlabel)
    else:
        plt.xlabel(dist_xvg.xvg_xlabel)
    plt.ylabel(dist_xvg.xvg_ylabel)
    plt.title(dist_xvg.xvg_title)
    plt.legend()
    if figout != None:
        plt.savefig("hbond_dist_" + figout, dpi=300)
    if not noshow:
        plt.show()

    plt.clf()
    ang_time = np.array(ang_xvg.xvg_columns[0]).astype(float)
    ang_time = ang_time * xshrink
    for i, hbond_name in enumerate(hbond_names):
        ang_line = np.array(ang_xvg.xvg_columns[i + 2]).astype(float)
        plt.plot(ang_time, ang_line, label=hbond_name)
    # plt.plot(ang_time, [30 for d in dist_time], "grey")
    if xlabel != None:
        plt.xlabel(xlabel)
    else:
        plt.xlabel(dist_xvg.xvg_xlabel)
    plt.ylabel(ang_xvg.xvg_ylabel)
    plt.title(ang_xvg.xvg_title)
    plt.legend()
    if figout != None:
        plt.savefig("hbond_angle_" + figout, dpi=300)
    if not noshow:
        plt.show()

    return dist_ave_std, ang_ave_std


def hbond(
    xpmfile: str = "",
    ndxfile: str = "",
    grofile: str = "",
    select: list = [],
    noshow: bool = False,
    figout: str = None,
    csv: str = None,
    hnf: str = None,
    genscript: bool = False,
    calc_distance_angle: bool = False,
    distancefile: str = None,
    anglefile: str = None,
    xshrink: float = 1.0,
    xlabel: str = None,
    set_operation: str = None,
) -> None:
    """
    hbond: a function to figure out hbond information, occupancy and occupancy
    map, occupancy table.
    This function will read one gro file, one hbmap.xpm, one hbond.ndx, and
    return some hbond information.

    :parameters:
        gro_file: the gro file which saved coordinates to pare atom names
        hbmap.xpm: the hbond occupancy map file generated by 'gmx hbond'
        hbond.ndx: the index file generated by 'gmx hbond'

    :return:
        None
    """

    for suffix, file in zip([".xpm", ".ndx", ".gro"], [xpmfile, ndxfile, grofile]):
        if file == None:
            logging.error("You have to specify the {} file for input".format(suffix))
            sys.exit()
        if not os.path.exists(file):
            logging.error("no {} in current dirrectory. ".format(file))
            sys.exit()
        if file[-4:] != suffix:
            logging.error(
                "only accept {} file with suffix {}".format(suffix.strip("."), suffix)
            )
            sys.exit()

    ## parse set operation list
    if set_operation != None:
        set_id_list = []
        if set_operation.startswith("AND"):
            set_flag = "AND"
            set_strs = set_operation[3:].strip().split()
        elif set_operation.startswith("OR"):
            set_flag = "OR"
            set_strs = set_operation[2:].strip().split()
        for set_str in set_strs:
            items = set_str.split(",")
            for item in items:
                if "-" in item:
                    set_id_list += [
                        i
                        for i in range(
                            int(item.split("-")[0]), int(item.split("-")[1]) + 1
                        )
                    ]
                else:
                    set_id_list.append(int(item))

    ## parse ndx file to get index of hydrogen bonds
    donor_ndxs, hydrogen_ndxs, acceptor_ndxs = [], [], []
    with open(ndxfile, "r") as fo:
        lines = [line for line in fo.readlines() if line.strip()]
    lines.reverse()
    for line in lines:
        if "hbonds_" in line:
            break
        items = line.split()
        donor_ndxs.append(int(items[0].strip()))
        hydrogen_ndxs.append(int(items[1].strip()))
        acceptor_ndxs.append(int(items[2].strip()))
    donor_ndxs.reverse()
    hydrogen_ndxs.reverse()
    acceptor_ndxs.reverse()
    if len(donor_ndxs) != len(hydrogen_ndxs) != len(acceptor_ndxs):
        logging.error("wrong length in donor, hydrogen, acceptor indexs")
        sys.exit()

    ## read the gro file and parse atom names
    if hnf == None:
        hnf = (
            "d_resname(d_resnum)@d_atomname(d_atomnum)->h_atomname("
            + "h_atomnum)...a_resname(a_resnum)@a_atomname(a_atomnum)"
        )
    features = [
        "d_resname",
        "d_resnum",
        "d_atomname",
        "d_atomnum",
        "h_resname",
        "h_resnum",
        "h_atomname",
        "h_atomnum",
        "a_resname",
        "a_resnum",
        "a_atomname",
        "a_atomnum",
    ]
    for feature in features:
        if feature in hnf or hnf == "number" or hnf == "id":
            break
    else:
        logging.warning(
            "no key feature detected in your specified hbond name "
            + "format, use the default format"
        )
        hnf = (
            "d_resname(d_resnum)@d_atomname(d_atomnum)->h_atomname("
            + "h_atomnum)...a_resname(a_resnum)@a_atomname(a_atomnum)"
        )

    donor_names, hydrogen_names, acceptor_names = [], [], []
    with open(grofile, "r") as fo:
        lines = fo.readlines()[2:-1]
    for ind in donor_ndxs:
        line = lines[ind - 1]
        res_num = line[:5].strip()
        res_name = line[5:10].strip()
        atom_name = line[10:15].strip()
        atom_num = line[15:20].strip()
        # name = f"{res_name}({res_num})@{atom_name}({atom_num})"
        # donor_names.append(name)
        donor_names.append([res_name, res_num, atom_name, atom_num])
    for ind in hydrogen_ndxs:
        line = lines[ind - 1]
        res_num = line[:5].strip()
        res_name = line[5:10].strip()
        atom_name = line[10:15].strip()
        atom_num = line[15:20].strip()
        # name = f"{res_name}({res_num})@{atom_name}({atom_num})"
        # hydrogen_names.append(name)
        hydrogen_names.append([res_name, res_num, atom_name, atom_num])
    for ind in acceptor_ndxs:
        line = lines[ind - 1]
        res_num = line[:5].strip()
        res_name = line[5:10].strip()
        atom_name = line[10:15].strip()
        atom_num = line[15:20].strip()
        # name = f"{res_name}({res_num})@{atom_name}({atom_num})"
        # acceptor_names.append(name)
        acceptor_names.append([res_name, res_num, atom_name, atom_num])
    if (
        len(donor_names)
        != len(hydrogen_names)
        != len(acceptor_names)
        != len(acceptor_ndxs)
    ):
        logging.error("wrong length in donor, hydrogen, acceptor names")
        sys.exit()
    # hbond_names = [
    #     f'{donor_names[i]}->{hydrogen_names[i].split("@")[1]}...{acceptor_names[i]}'
    #     for i in range(len(donor_names))]
    hbond_names = []
    for i in range(len(donor_names)):
        hbond_name = hnf.replace("d_resname", donor_names[i][0])
        hbond_name = hbond_name.replace("d_resnum", donor_names[i][1])
        hbond_name = hbond_name.replace("d_atomname", donor_names[i][2])
        hbond_name = hbond_name.replace("d_atomnum", donor_names[i][3])
        hbond_name = hbond_name.replace("h_resname", hydrogen_names[i][0])
        hbond_name = hbond_name.replace("h_resnum", hydrogen_names[i][1])
        hbond_name = hbond_name.replace("h_atomname", hydrogen_names[i][2])
        hbond_name = hbond_name.replace("h_atomnum", hydrogen_names[i][3])
        hbond_name = hbond_name.replace("a_resname", acceptor_names[i][0])
        hbond_name = hbond_name.replace("a_resnum", acceptor_names[i][1])
        hbond_name = hbond_name.replace("a_atomname", acceptor_names[i][2])
        hbond_name = hbond_name.replace("a_atomnum", acceptor_names[i][3])
        hbond_names.append(hbond_name)

    ## parse xpmfile
    xpm = XPM(xpmfile)
    if xpm.xpm_height < len(hbond_names):
        logging.warning(
            "height of xpm ({}) in {} is not equal to number of hbond ({}) in {}, removed the excess in xpm".format(
                xpm.xpm_height, xpmfile, len(hydrogen_names), ndxfile
            )
        )
        gap = xpm.xpm_height - len(hbond_names)
        xpm.xpm_height -= gap
        xpm.xpm_datalines = xpm.xpm_datalines[gap:]
        xpm.xpm_yaxis = xpm.xpm_yaxis[gap:]
    xpm.xpm_datalines.reverse()
    xpm.xpm_yaxis.reverse()

    ## get occupancy
    occupancy, xpm_datamatrix = [], []
    for dataline in xpm.xpm_datalines:
        dot_list = []
        for i in range(
            0, xpm.xpm_width * xpm.xpm_char_per_pixel, xpm.xpm_char_per_pixel
        ):
            dot_list.append(xpm.chars.index(dataline[i : i + xpm.xpm_char_per_pixel]))
        xpm_datamatrix.append(dot_list)
        occupancy.append(sum(dot_list) / (1.0 * len(dot_list)))

    # set operation
    if set_operation != None:
        set_line = []
        for s in set_id_list:
            if s >= len(xpm_datamatrix):
                logging.error(
                    f"hbond id in set_operation out of range, it should be in 0 to {len(xpm_datamatrix)}(not include)"
                )
                sys.exit()
        for x in range(xpm.xpm_width):
            res = xpm_datamatrix[set_id_list[0]][x]
            for id in set_id_list:
                if set_flag == "AND":
                    res = res and xpm_datamatrix[id][x]
                elif set_flag == "OR":
                    res = res or xpm_datamatrix[id][x]
                else:
                    logging.error("Wrong set operation, only support AND and OR")
            set_line.append(res)

    ## deal with the selection
    # select = [5]
    if select == []:
        select = [i for i in range(len(hbond_names))]
    else:
        occupancy = [occupancy[i] for i in select]
        xpm_datamatrix = [xpm_datamatrix[i] for i in select]
    if hnf == "number":
        hbond_names = [str(i) for i in range(len(select))]
    elif hnf == "id":
        hbond_names = [str(i) for i in select]
    else:
        hbond_names = [hbond_names[i] for i in select]

    if set_operation != None:
        hbond_names.append(set_operation)
        occupancy.append(sum(set_line) / (1.0 * len(set_line)))
        xpm_datamatrix.append(set_line)
        select.append(-1)

    ## draw map
    if xshrink != None and xshrink != 1.0:
        xpm.xpm_xaxis = [x * xshrink for x in xpm.xpm_xaxis]
        xpm.xpm_xlabel = (xlabel, xpm.xpm_xlabel)[xlabel == None]
    plt.figure()
    cmap = mcolors.ListedColormap(["white", "#F94C66"])
    hbond = plt.pcolormesh(
        xpm.xpm_xaxis,
        [i for i in range(len(select))],
        xpm_datamatrix,
        cmap=cmap,
        shading="auto",
    )
    if len(select) == 1:
        hbond = plt.imshow(xpm_datamatrix, cmap=cmap, aspect="auto")
        # TODO: xticks
    cb = plt.colorbar(hbond, orientation="horizontal", fraction=0.03)
    cb.set_ticks([0.25, 0.75])
    cb.set_ticklabels(["None", "Present"])
    for i in range(1, len(select)):
        plt.hlines(i - 0.5, min(xpm.xpm_xaxis), max(xpm.xpm_xaxis), colors="white")
    plt.title(xpm.xpm_title)
    plt.xlabel(xpm.xpm_xlabel)
    plt.ylabel(xpm.xpm_ylabel)
    if xpm.xpm_height <= 20 and hnf == "number":
        plt.yticks([i for i in range(len(select))], [i for i in range(len(select))])
    elif xpm.xpm_height <= 20 and hnf == "id":
        plt.yticks([i for i in range(len(select))], select)
    elif xpm.xpm_height <= 20 and hnf != "number" and hnf != "id":
        plt.yticks([i for i in range(len(select))], hbond_names)
    else:
        logging.warning("DIT can only show hbond name when hbond number <= 20 !")
    plt.tight_layout()
    if figout != None:
        plt.savefig(figout, dpi=300)
    if not noshow:
        plt.show()

    if set_operation != None and (genscript or calc_distance_angle):
        logging.error(
            "you are not surposed to specify set_operation and genscript or calc_distance_angle together, remove set_operation!"
        )
        sys.exit()
    if genscript:
        logging.warning(
            "!IMPORTANT! remember to set 'merge' option of 'gmx hbond' to 'no' if you wanna to calculate average hbond distance and angle !"
        )
        gen_distang_script(donor_ndxs, hydrogen_ndxs, acceptor_ndxs, select)
    dist_ave_std, ang_ave_std = [], []
    if calc_distance_angle and distancefile != None and anglefile != None:
        dist_ave_std, ang_ave_std = calculate_distance_angle(
            xpm_datamatrix,
            hbond_names,
            distancefile,
            anglefile,
            figout,
            noshow,
            xshrink,
            xlabel,
        )

    ## show table
    print("-" * 115)
    print(
        "{:<2} {:<50} {:>14} {:>14}".format(
            "id", "donor->hydrogen...acceptor", "occupancy(%)", "Present/Frames"
        ),
        end="",
    )
    if len(dist_ave_std) != 0 and len(ang_ave_std) != 0:
        print(" {:>15} {:>15}".format("Distance (nm)", "Angle (°)"))
    else:
        print()
    print("-" * 115)
    for i in range(len(hbond_names)):
        print(
            "{:<2d} {:<50} {:>14.2f} {:>6d}/{:<7}".format(
                select[i],
                hbond_names[i],
                occupancy[i] * 100.0,
                np.sum(xpm_datamatrix[i]),
                len(xpm_datamatrix[i]),
            ),
            end="",
        )
        if len(dist_ave_std) != 0 and len(ang_ave_std) != 0:
            print(
                "  {:>6.4f} ± {:<6.4f} {:>6.2f} ± {:<6.2f}".format(
                    dist_ave_std[i][0],
                    dist_ave_std[i][1],
                    ang_ave_std[i][0],
                    ang_ave_std[i][1],
                )
            )
        else:
            print()
    print("-" * 115)
    if csv != None:
        with open(csv, "w") as fo:
            fo.write(
                "{},{},{},{}".format(
                    "id", "donor->hydrogen...acceptor", "occupancy(%)", "Present/Frames"
                )
            )
            if len(dist_ave_std) != 0 and len(ang_ave_std) != 0:
                fo.write(
                    ",{},{},{},{}\n".format(
                        "Distance ave (nm)",
                        "Distance std (nm)",
                        "Angle ave (°)",
                        "Angle std (°)",
                    )
                )
            else:
                fo.write("\n")
            for i in range(len(hbond_names)):
                fo.write(
                    "{},{},{:.2f},{:>d}/{}".format(
                        select[i],
                        hbond_names[i],
                        occupancy[i] * 100.0,
                        np.sum(xpm_datamatrix[i]),
                        len(xpm_datamatrix[i]),
                    )
                )
                if len(dist_ave_std) != 0 and len(ang_ave_std) != 0:
                    fo.write(
                        ",{:>6.4f},{:<6.4f},{:>6.2f},{:<6.2f}\n".format(
                            dist_ave_std[i][0],
                            dist_ave_std[i][1],
                            ang_ave_std[i][0],
                            ang_ave_std[i][1],
                        )
                    )
                else:
                    fo.write("\n")


def hbond_call_functions(arguments: list = []):
    """call functions according to arguments"""

    if len(arguments) == 0:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(
        description="hbond command process the hbond infos"
    )
    parser.add_argument("-f", "--input", help="gro file for input")
    parser.add_argument("-n", "--index", help="hbond ndx file for input")
    parser.add_argument("-m", "--map", help="hbond map file for input")
    parser.add_argument(
        "-c", "--select", nargs="+", help="to select row of data, like: 1 2-4 6"
    )
    parser.add_argument("-o", "--output", help="figure name for output")
    parser.add_argument("-csv", "--csv", help="store table info into csv file")
    parser.add_argument(
        "-ns",
        "--noshow",
        action="store_true",
        help="whether not to show picture, useful on computer without gui",
    )
    parser.add_argument(
        "-hnf",
        "--hbond_name_format",
        help="define the hbond name format by user! Each atom has four"
        + " features: resname, resnum, atomname, atomnum. Distinguish "
        + "donor, hydrogen, acceptor by adding one prefix to each feature,"
        + " like: d_resname, a_resnum, h_atomname. \nSo you may able to "
        + "define hbond name style by: 'd_resname(d_resnum)@d_atomname(d_"
        + "atomnum)->h_atomname(h_atomnum)...a_resname(a_resnum)@a_atomn"
        + "ame(a_atomnum)' which is the default style,  or also you could"
        + " specify 'd_atomname@h_atomname...a_atomname' or some format you "
        + "would like. \nOr you could just set the hnf to be 'number' or 'id'",
    )
    parser.add_argument(
        "-genscript",
        "--genscript",
        action="store_true",
        help="whether to generate scripts for calculating distance and angle of hbonds",
    )
    parser.add_argument(
        "-cda",
        "--calc_distance_angle",
        action="store_true",
        help="whether to calculate distance and angle of hbonds from distance xvg file and angle xvg file",
    )
    parser.add_argument(
        "-distancefile", "--distancefile", help="distance file of hbonds for input"
    )
    parser.add_argument(
        "-anglefile", "--anglefile", help="angle file of hbonds for input"
    )
    parser.add_argument(
        "-xs", "--xshrink", type=float, help="modify x-axis by multipling xshrink"
    )
    parser.add_argument("-x", "--xlabel", type=str, help="the xlabel of figure")
    parser.add_argument(
        "-so",
        "--set_operation",
        type=str,
        help="use AND or OR to operate different hbonds. eg. -so AND1-2,4,7  -so OR0,4,6-8",
    )

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        exit()

    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        exit()
    if len(arguments) == 2:
        logging.error("no parameters, type 'dit <command> -h' for more infos.")
        exit()

    args = parser.parse_args(arguments[2:])
    select = []
    if args.select != None:
        for c in args.select:
            if "-" in c.strip("-"):
                select += [i for i in range(int(c.split("-")[0]), int(c.split("-")[1]))]
            else:
                select.append(int(c))

    if method == "hbond":
        grofile = args.input
        ndxfile = args.index
        xpmfile = args.map
        noshow = args.noshow
        figout = args.output
        csv = args.csv
        hnf = args.hbond_name_format
        genscript = args.genscript
        calc_distance_angle = args.calc_distance_angle
        distancefile = args.distancefile
        anglefile = args.anglefile
        xshrink = args.xshrink
        xlabel = args.xlabel
        set_operation = args.set_operation
        hbond(
            xpmfile,
            ndxfile,
            grofile,
            select,
            noshow,
            figout,
            csv,
            hnf,
            genscript,
            calc_distance_angle,
            distancefile,
            anglefile,
            xshrink,
            xlabel,
            set_operation,
        )
    else:
        logging.error("unknown method {}".format(method))
        exit()

    logging.info("May you good day !")


def main():
    hbond_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()



"""
groParser module is part of DuIvyTools for parsing the gro file generated by GROMACS.

Written by DuIvy and provided to you by GPLv3 license.
"""

import os
import sys

base = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))
if base not in sys.path:
    sys.path.insert(0, base)

from utils import log


class Atom(object):
    """Atom class for parsing atom line in gro file"""

    def __init__(self, line: str) -> None:
        self.res_id = int(line[:5].strip())
        self.res_name = line[5:10].strip()
        self.atom_name = line[10:15].strip()
        self.atom_id = int(line[15:20].strip())
        self.coor_x = float(line[20:28].strip())
        self.coor_y = float(line[28:36].strip())
        self.coor_z = float(line[36:44].strip())
        self.coor = (self.coor_x, self.coor_y, self.coor_z)
        if len(line) > 45:
            self.velocity_x = float(line[44:52].strip())
            self.velocity_y = float(line[52:60].strip())
            self.velocity_z = float(line[60:68].strip())
        else:
            self.velocity_x = None
            self.velocity_y = None
            self.velocity_z = None
        self.velocity = (self.velocity_x, self.velocity_y, self.velocity_z)

    def __str__(self) -> str:
        output: str = f"""{self.res_id:>5}{self.res_name:<5}{self.atom_name:>5}"""
        output += f"""{self.atom_id:>5}{self.coor_x:>8}"""
        output += f"""{self.coor_y:>8}{self.coor_z:>8}"""
        return output


class GRO(log):
    """GRO class for parsing gro file"""

    def __init__(self, grofile: str, new_file: bool = False) -> None:
        self.frame_num: int = 0
        self.atom_number: int = 0
        self.frames: list[list[Atom]] = []
        self.notes: list[str] = []
        self.box_coors: list[tuple] = []
        self.grofile: str = grofile

        if not new_file and grofile:
            if not os.path.exists(grofile):
                self.error(f"No {grofile} detected ! check it !")
            if grofile[-4:] != ".gro":
                self.error(
                    f"you must specify a file with suffix .gro, instead of {grofile}"
                )
            with open(grofile, "r") as fo:
                lines = fo.readlines()
            try:
                self.atom_number = int(lines[1].strip())
            except:
                self.error("The second line of gro file must be Int number")
            self.frame_num = len(lines) // (self.atom_number + 3)
            for f in range(self.frame_num):
                atom_list: list[Atom] = []
                for line in lines[
                    f * (self.atom_number + 3)
                    + 2 : (f + 1) * (self.atom_number + 3)
                    - 1
                ]:
                    atom_list.append(Atom(line))
                self.frames.append(atom_list)
                self.notes.append(lines[f * (self.atom_number + 3)])
                coor_line = lines[(f + 1) * (self.atom_number + 3) - 1].strip().split()
                self.box_coors.append(tuple([float(c) for c in coor_line]))

    def get_time_info(self):
        pass



"""
XPM module is part of DuIvyTools library, which is a tool for analysis and 
visualization of GROMACS result files. This module is written by CharlesHahn.

XPM module process xpm files, including getting information from xpm, 
visualization and data convertion. 

This module requires Numpy, Scipy, Matplotlib and argparse. 

This module contains:
    class XPM
    function xpm_combine
    ......

This file is provided to you under GPLv3 License"""

## TODO: data convertion and process in XPM class

import os
import sys
import math
import argparse
import numpy as np
from scipy.interpolate import interp2d
from scipy.ndimage import gaussian_filter
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoLocator, FormatStrFormatter
from matplotlib import patches
import logging


class XPM(object):
    """class XPM was defined to process xpm files

    Attributes:
    xpmfile: the name of xpm file
    xpm_title: the title of xpm file
    xpm_legend: the legend of xpm file
    xpm_type: two type of xpm file, "Continuous", "Discrete"
    xpm_xlabel: the x-label of xpm file
    xpm_ylabel: the y-label of xpm file
    xpm_width: the width of xpm file
    xpm_height: the height of xpm file
    xpm_color_num: the number of colors in xpm file
    xpm_char_per_pixel: the number of chars to represent a pixel of xpm figure
    chars: a list to store the chars which reprent pixels in xpm figure
    colors: a list to store the colors in xpm figure
    notes: a list to store the notes (comments to chars and colors) in xpm figure
    colors_rgb: a list of colors in rgb type
    xpm_xaxis: a list to store x-axis values
    xpm_yaxis: a list to store y-axis values
    xpm_datalines: a list to store each line of xpm figure content

    Methods:
    get_scatter_data(self): convert xpm figure content to scatter data
        return:
            scatter_x, scatter_y, x, y, v
    xpm2csv(self, outcsv: str = ""): convert xpm data to csv data
            outcsv: the output csv file name
    xpm2gpl(self, outgpl: str = ""): convert xpm data to gnuplot script
            outgpl: the output gpl file name
    draw_origin(self, IP: bool = False, outputpng: str = None, noshow: bool = False): draw xpm by plt.imshow()
            IP : whether to do interpolation
            outputpng : the output figure file name
            noshow: whether not to show figure, useful for computer without GUI
    draw_pcm(self, IP: bool = False, outputpng: str = None, noshow: bool = False): draw xpm by plt.pcolormesh()
            IP : whether to do interpolation
            outputpng : the output figure file name
            noshow: whether not to show figure, useful for computer without GUI
    draw_3D(self, IP: bool = False, outputpng: str = None, noshow: bool = False): draw xpm in 3D style
            IP : whether to do interpolation
            outputpng : the output figure file name
            noshow: whether not to show figure, useful for computer without GUI
    """

    def __init__(
        self,
        xpmfile: str = None,
        xlabel: str = None,
        ylabel: str = None,
        title: str = None,
        xshrink: float = 1.0,
    ):
        """
        read xpm file and save infos to class xpm

        :parameters:
            xpmfile: the xpm file name
            xlabel: specify the xlabel of figure
            ylabel: specify the ylabel of figure
            title: specify the title of figure
            xshrink: specify the factor for multiplication of x-axis
        """

        ## check parameters
        if xpmfile == None:
            logging.error("no input xpm file detected")
            sys.exit()
        if not os.path.exists(xpmfile):
            logging.error("no {} in current directory".format(xpmfile))
            sys.exit()
        if xpmfile[-4:] != ".xpm":
            logging.error("specify a xpm file with suffix xpm")
            sys.exit()
        if xshrink == None:
            xshrink = 1.0

        self.xpmfile = xpmfile
        self.xpm_title = ""
        self.xpm_legend = ""
        self.xpm_type = ""
        self.xpm_xlabel = ""
        self.xpm_ylabel = ""
        self.xpm_width = 0
        self.xpm_height = 0
        self.xpm_color_num = 0
        self.xpm_char_per_pixel = 0
        self.chars = []
        self.colors = []
        self.notes = []
        self.colors_rgb = []
        self.xpm_xaxis = []
        self.xpm_yaxis = []
        self.xpm_datalines = []

        ## read xpm file
        with open(xpmfile, "r") as fo:
            lines = [line.strip() for line in fo.readlines()]

        ## parse content of xpmfile
        flag_4_code = 0  ## means haven't detected yet
        for line in lines:
            ## find the 4 code line and parse
            if flag_4_code == 1:  ## means this line is code4 line
                flag_4_code = 2  ## means have detected
                code4 = [int(c) for c in line.strip().strip(",").strip('"').split()]
                self.xpm_width, self.xpm_height = code4[0], code4[1]
                self.xpm_color_num, self.xpm_char_per_pixel = code4[2], code4[3]
                continue
            elif (flag_4_code == 0) and line.startswith("static char"):
                flag_4_code = 1  ## means next line is code4 line
                continue

            ## parse comments and axis parts
            if line.startswith("/* x-axis"):
                self.xpm_xaxis += [float(n) for n in line.strip().split()[2:-1]]
                continue
            elif line.startswith("/* y-axis"):
                self.xpm_yaxis += [float(n) for n in line.strip().split()[2:-1]]
                continue
            elif line.startswith("/* title"):
                self.xpm_title = line.strip().split('"')[1]
                continue
            elif line.startswith("/* legend"):
                self.xpm_legend = line.strip().split('"')[1]
                continue
            elif line.startswith("/* x-label"):
                self.xpm_xlabel = line.strip().split('"')[1]
                continue
            elif line.startswith("/* y-label"):
                self.xpm_ylabel = line.strip().split('"')[1]
                continue
            elif line.startswith("/* type"):
                self.xpm_type = line.strip().split('"')[1]
                continue

            items = line.strip().split()
            ## for char-color-note part
            if len(items) == 7 and items[1] == "c":
                if len(items[0].strip('"')) == self.xpm_char_per_pixel:
                    self.chars.append(items[0].strip('"'))
                    self.colors.append(items[2])
                    self.notes.append(items[5].strip('"'))
                ## deal with blank
                if len(items[0].strip('"')) < self.xpm_char_per_pixel:
                    logging.warning("space in char of line : {}".format(line))
                    char_item = items[0].strip('"')
                    self.chars.append(
                        char_item + " " * (self.xpm_char_per_pixel - len(char_item))
                    )
                    self.colors.append(items[2])
                    self.notes.append(items[5].strip('"'))
                continue

            ## for content part
            if line.strip().startswith('"') == 1 and (
                len(line.strip().strip(",").strip('"'))
                == self.xpm_width * self.xpm_char_per_pixel
            ):
                self.xpm_datalines.append(line.strip().strip(",").strip('"'))

        ## check infos
        if len(self.chars) != len(self.colors) != len(self.notes) != self.xpm_color_num:
            logging.error("length of chars, colors, notes != xpm_color_num")
            logging.info(
                "chars : {}, colors : {}, notes : {}, xpm_color_num : {}".format(
                    len(self.chars),
                    len(self.colors),
                    len(self.notes),
                    self.xpm_color_num,
                )
            )
            sys.exit()

        if len(self.xpm_datalines) != self.xpm_height:
            logging.error(
                "rows of data ({}) is not equal to xpm height ({}), check it !".format(
                    len(self.xpm_datalines), self.xpm_height
                )
            )
            sys.exit()
        if (
            len(self.xpm_xaxis) != self.xpm_width
            and len(self.xpm_xaxis) != self.xpm_width + 1
        ):
            logging.error(
                "length of x-axis ({}) != xpm width ({}) or xpm width +1".format(
                    len(self.xpm_xaxis), self.xpm_width
                )
            )
            sys.exit()
        if (
            len(self.xpm_yaxis) != self.xpm_height
            and len(self.xpm_yaxis) != self.xpm_height + 1
        ):
            logging.error(
                "length of y-axis ({}) != xpm height ({}) or xpm height +1".format(
                    len(self.xpm_yaxis), self.xpm_height
                )
            )
            sys.exit()

        if len(self.xpm_xaxis) == self.xpm_width + 1:
            self.xpm_xaxis = [
                (self.xpm_xaxis[i - 1] + self.xpm_xaxis[i]) / 2.0
                for i in range(1, len(self.xpm_xaxis))
            ]
            logging.warning(
                "length of x-axis is 1 more than xpm width, use intermediate value for instead. "
            )
        if len(self.xpm_yaxis) == self.xpm_height + 1:
            self.xpm_yaxis = [
                (self.xpm_yaxis[i - 1] + self.xpm_yaxis[i]) / 2.0
                for i in range(1, len(self.xpm_yaxis))
            ]
            logging.warning(
                "length of y-axis is 1 more than xpm height, use intermediate value for instead. "
            )

        ## hex color to rgb values
        for color in self.colors:
            r = int(color[1:3], 16)
            g = int(color[3:5], 16)
            b = int(color[5:7], 16)
            self.colors_rgb.append([r, g, b])

        ## for xlabel and ylabel modification
        if xlabel != None:
            self.xpm_xlabel = xlabel
        if ylabel != None:
            self.xpm_ylabel = ylabel
        if title != None:
            self.xpm_title = title
        if xshrink != None:
            self.xpm_xaxis = [x * float(xshrink) for x in self.xpm_xaxis]

        ## the read order of pixels is from top to bottom
        ## but the y-axis is from bottom to top, so reverse() is important !
        self.xpm_yaxis.reverse()

        logging.info("all data has been read from {} successfully.".format(xpmfile))

    def get_scatter_data(self) -> tuple:
        """convert xpm data to scatter data"""

        ## parse xpm_data into x, y, v
        x, y, v = [], [], []
        scatter_x, scatter_y = [], []
        for l in range(len(self.xpm_datalines)):
            for i in range(
                0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
            ):
                v.append(
                    float(
                        self.notes[
                            self.chars.index(
                                self.xpm_datalines[l][i : i + self.xpm_char_per_pixel]
                            )
                        ]
                    )
                )
                x.append(self.xpm_xaxis[int(i / self.xpm_char_per_pixel)])
                y.append(self.xpm_yaxis[l])

        ## parse x, y, v into scatter_x, scatter_y
        v_max = max(v)
        scatter_weight = 1
        for i in range(len(v)):
            count = round((v_max - v[i]) * scatter_weight)
            for _ in range(count):
                scatter_x.append(x[i])
                scatter_y.append(y[i])

        return scatter_x, scatter_y, x, y, v

    def xpm2csv(self, outcsv: str = "") -> None:
        """convert xpm file to csv file
        outcsv: the name of output csv file
        """

        if outcsv == None:
            outcsv = ""
        if outcsv == "":
            outcsv = self.xpmfile[:-4] + ".csv"
        if outcsv[-4:] != ".csv":
            logging.error("specify a output file with suffix csv")
            sys.exit()
        if os.path.exists(outcsv):
            logging.error("{} already in current directory".format(outcsv))
            sys.exit()

        if self.xpm_type != "Continuous":
            logging.error("can not extract data from xpm whose type is not Continuous")
            sys.exit()

        ## only x, y, v values are needed
        _, _, x, y, v = self.get_scatter_data()
        if len(x) != len(y) != len(v):
            logging.error("wrong in length of x, y, v")
            sys.exit()
        ## write results
        with open(outcsv, "w") as fo:
            x_title = (self.xpm_xlabel, "x-axis")[len(self.xpm_xlabel) == 0]
            y_title = (self.xpm_ylabel, "y-axis")[len(self.xpm_ylabel) == 0]
            z_title = (self.xpm_legend, "value")[len(self.xpm_legend) == 0]
            fo.write("{},{},{}\n".format(x_title, y_title, z_title))
            for i in range(len(x)):
                fo.write("{:.6f},{:.6f},{:.6f}\n".format(x[i], y[i], v[i]))
        logging.info("extract data from {} successfully".format(self.xpmfile))
        logging.info("data are saved into {}".format(outcsv))

    def xpm2dat(self, outdat: str = "") -> None:
        """convert xpm file to dat file, N*N
        outdat: the name of output dat file
        """

        if outdat == None:
            outadat = ""
        if outdat == "":
            outdat = self.xpmfile[:-4] + ".dat"
        if outdat[-4:] != ".dat":
            logging.error("specify a output file with suffix dat")
            sys.exit()
        if os.path.exists(outdat):
            logging.error("{} already in current directory".format(outdat))
            sys.exit()

        if self.xpm_type != "Continuous":
            logging.error("can not extract data from xpm whose type is not Continuous")
            sys.exit()

        ## write results
        with open(outdat, "w") as fo:
            x_title = (self.xpm_xlabel, "x-axis")[len(self.xpm_xlabel) == 0]
            y_title = (self.xpm_ylabel, "y-axis")[len(self.xpm_ylabel) == 0]
            z_title = (self.xpm_legend, "value")[len(self.xpm_legend) == 0]
            fo.write(
                "# first line: {}; second line (bottom to top): {}; Values :{}\n".format(
                    x_title, y_title, z_title
                )
            )
            fo.write(",".join(["{:.4f}".format(item) for item in self.xpm_xaxis]))
            fo.write("\n")
            true_yaxis = ["{:.4f}".format(item) for item in self.xpm_yaxis]
            true_yaxis.reverse()
            fo.write(",".join(true_yaxis))
            fo.write("\n")
            for l in range(len(self.xpm_datalines)):
                value = []
                for i in range(
                    0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
                ):
                    value.append(
                        float(
                            self.notes[
                                self.chars.index(
                                    self.xpm_datalines[l][
                                        i : i + self.xpm_char_per_pixel
                                    ]
                                )
                            ]
                        )
                    )
                fo.write(",".join(["{:.6f}".format(v) for v in value]))
                fo.write("\n")
        logging.info("extract data from {} successfully".format(self.xpmfile))
        logging.info(
            "{} * {} matrix data are saved into {}".format(
                self.xpm_width, self.xpm_height, outdat
            )
        )

    def xpm2gpl(self, outgpl: str = "") -> None:
        """convert xpm file to gnuplot script
        outgpl: the name of output gnuplot script
        """

        ## check files
        if outgpl == None:
            outgpl = ""
        if outgpl == "":
            outgpl = self.xpmfile[:-4] + ".gpl"
        if os.path.exists(outgpl):
            logging.error("{} already in current directory".format(outgpl))
            sys.exit()
        outpng = self.xpmfile[:-4] + ".png"

        ## write gnuplot scripts
        gpl_lines = "set term png\n"
        gpl_lines += """set output "{}" \n""".format(outpng)
        gpl_lines += "unset colorbox\n"
        pal_line = "set pal defined("
        for index, color in enumerate(self.colors):
            pal_line += """{} "{}",""".format(index, color)
        pal_line = pal_line.strip(",") + ")"
        gpl_lines += pal_line + "\n\n"
        ## add data lines
        gpl_lines += "$data << EOD\n"
        for l in range(len(self.xpm_datalines)):
            for i in range(
                0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
            ):
                value = self.chars.index(
                    self.xpm_datalines[l][i : i + self.xpm_char_per_pixel]
                )
                gpl_lines += "{:.6f} {:.6f} {:.6f}\n".format(
                    self.xpm_xaxis[int(i / self.xpm_char_per_pixel)],
                    self.xpm_yaxis[l],
                    value,
                )
        gpl_lines += "EOD\n\n"
        ## add tail part of gpl file
        gpl_lines += "#set tmargin at screen 0.95\n"
        gpl_lines += "#set bmargin at screen 0.20\n"
        gpl_lines += "#set rmargin at screen 0.85\n"
        y_posi = 0.92
        for index, note in enumerate(self.notes):
            label_line = """#set label "{:10}" at screen 0.85,{:.2f} left textcolor rgb "{}"\n""".format(
                note, y_posi, self.colors[index]
            )
            y_posi -= 0.10
            gpl_lines += label_line
        gpl_lines += """set term pngcairo enhanced truecolor font "Arial,85" """
        gpl_lines += """fontscale 1 linewidth 20 pointscale 5 size 10000,6000\n"""
        gpl_lines += "set tics out nomirror;\n"
        gpl_lines += "set key out reverse Left spacing 2 samplen 1/2\n"
        gpl_lines += """set title "{}"\n""".format(self.xpm_title)
        gpl_lines += """set xlabel "{}"; set ylabel "{}";\n""".format(
            self.xpm_xlabel, self.xpm_ylabel
        )
        if self.xpm_type == "Continuous":
            gpl_lines += """plot [{:.2f}:{:.2f}] [{:.2f}:{:.2f}] $data u 1:2:3 w imag notit, \\\n""".format(
                math.floor(min(self.xpm_xaxis) * 10.0) / 10.0 - 0.1,
                math.ceil(max(self.xpm_xaxis) * 10.0) / 10.0 + 0.1,
                math.floor(min(self.xpm_yaxis) * 10.0) / 10.0 - 0.1,
                math.ceil(max(self.xpm_yaxis) * 10.0) / 10.0 + 0.1,
            )
        else:
            gpl_lines += """plot [{:.2f}:{:.2f}] [{:.2f}:{:.2f}] $data u 1:2:3 w imag notit, \\\n""".format(
                math.floor(min(self.xpm_xaxis) * 10.0) / 10.0 - 0.1,
                math.ceil(max(self.xpm_xaxis) * 10.0) / 10.0 + 0.1,
                math.floor(min(self.xpm_yaxis) * 10.0) / 10.0 - 0.6,
                math.ceil(max(self.xpm_yaxis) * 10.0) / 10.0 + 0.6,
            )

        for index, note in enumerate(self.notes):
            gpl_lines += """{} w p ps 3 pt 5 lc rgb "{}" t"{}", \\\n""".format(
                math.floor(min(self.xpm_yaxis)) - 1, self.colors[index], note
            )
        gpl_lines = gpl_lines.strip("\n").strip("\\").strip().strip(",")

        ## write gpl files
        with open(outgpl, "w") as fo:
            fo.write(gpl_lines + "\n")

        logging.info(
            "write gnuplot scripts {} from {} successfully".format(outgpl, self.xpmfile)
        )

    def draw_origin(
        self, IP: bool = False, outputpng: str = None, noshow: bool = False
    ) -> None:
        """draw xpm figure by plt.imshow
        IP : whether to interpolation
        outputpng: the name for figure output
        noshow: whether not to show figure, useful for PC without gui
        """

        ## check parameters
        if outputpng != None and os.path.exists(outputpng):
            logging.error("{} already in current directory".format(outputpng))
            sys.exit()

        # visualization of xpm
        plt.clf()
        if IP == False:
            img = []
            for line in self.xpm_datalines:
                rgb_line = []
                for i in range(
                    0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
                ):
                    rgb_line.append(
                        self.colors_rgb[
                            self.chars.index(line[i : i + self.xpm_char_per_pixel])
                        ]
                    )
                img.append(rgb_line)
            ## default of interpolation is nearest, set it to be none
            plt.imshow(img, aspect="auto", interpolation="none")

            if self.xpm_type != "Continuous":
                legend_patches = []
                for ind, note in enumerate(self.notes):
                    leg_patch = patches.Patch(color=self.colors[ind], label=note)
                    legend_patches.append(leg_patch)
                plt.legend(
                    handles=legend_patches,
                    bbox_to_anchor=(1.02, 1.00),
                    loc="upper left",
                    borderaxespad=0,
                )

        if IP == True:
            if self.xpm_type != "Continuous":
                logging.error("Only Continuous type xpm file can interpolation")
                sys.exit()
            ## show figure with interpolation
            imgIP = []
            for line in self.xpm_datalines:
                value_line = []
                for i in range(
                    0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
                ):
                    value_line.append(
                        float(
                            self.notes[
                                self.chars.index(line[i : i + self.xpm_char_per_pixel])
                            ]
                        )
                    )
                imgIP.append(value_line)

            im = plt.imshow(imgIP, cmap="jet", interpolation="bilinear", aspect="auto")
            plt.colorbar(im, label=self.xpm_legend)

        ## TODO: find a better way to solve problem of ticks
        ## set the ticks
        if self.xpm_height > 3 and self.xpm_width > 3:
            x_tick, y_tick = 3, 3
            xpm_xticks = ["{:.1f}".format(x) for x in self.xpm_xaxis]
            xpm_yticks = ["{:.1f}".format(y) for y in self.xpm_yaxis]
            if self.xpm_width < 100:
                x_tick = int(self.xpm_width / 3)
            elif self.xpm_width >= 100 and self.xpm_width < 1000:
                x_tick = int(self.xpm_width / 5)
            elif self.xpm_width > 500:
                x_tick = int(self.xpm_width / 10)
            if self.xpm_height < 100:
                y_tick = int(self.xpm_height / 3)
            elif self.xpm_height >= 100 and self.xpm_height < 1000:
                y_tick = int(self.xpm_height / 5)
            elif self.xpm_height > 500:
                y_tick = int(self.xpm_height / 10)
            if self.xpm_width / self.xpm_height > 10:
                y_tick = int(self.xpm_height / 2)
            if self.xpm_height / self.xpm_width > 10:
                x_tick = int(self.xpm_width / 2)
            plt.tick_params(axis="both", which="major")
            plt.xticks(
                [0]
                + [w for w in range(x_tick, self.xpm_width - int(x_tick / 2), x_tick)]
                + [self.xpm_width - 1],
                [xpm_xticks[0]]
                + [
                    xpm_xticks[w]
                    for w in range(x_tick, self.xpm_width - int(x_tick / 2), x_tick)
                ]
                + [xpm_xticks[-1]],
            )
            plt.yticks(
                [0]
                + [h for h in range(y_tick, self.xpm_height - int(y_tick / 2), y_tick)]
                + [self.xpm_height - 1],
                [xpm_yticks[0]]
                + [
                    xpm_yticks[h]
                    for h in range(y_tick, self.xpm_height - int(y_tick / 2), y_tick)
                ]
                + [xpm_yticks[-1]],
            )
        else:
            xpm_yticks = ["{:.0f}".format(y) for y in self.xpm_yaxis]
            plt.yticks([y for y in range(self.xpm_height)], xpm_yticks)

        ## set other infos in the figure
        plt.title(self.xpm_title)
        plt.xlabel(self.xpm_xlabel)
        plt.ylabel(self.xpm_ylabel)
        logging.info("Legend of this xpm figure -> {}".format(self.xpm_legend))
        plt.tight_layout()

        if outputpng != None:
            plt.savefig(outputpng, dpi=300)
        if noshow == False:
            plt.show()

    def draw_pcm(
        self, IP: bool = False, outputpng: str = None, noshow: bool = False
    ) -> None:
        """draw xpm figure by pcolormesh (with interpolation)
        IP : whether to interpolation
        outputpng: the name for figure output
        noshow: whether not to show figure, useful for PC without gui
        """

        ## check parameters
        if outputpng != None and os.path.exists(outputpng):
            logging.error("{} already in current directory".format(outputpng))
            sys.exit()

        if self.xpm_type != "Continuous":
            logging.error("Only Continuous type xpm file can interpolation")
            sys.exit()

        ## convert xpm_data to img (values)
        plt.clf()
        img = []
        for line in self.xpm_datalines:
            value_line = []
            for i in range(
                0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
            ):
                value_line.append(
                    float(
                        self.notes[
                            self.chars.index(line[i : i + self.xpm_char_per_pixel])
                        ]
                    )
                )
            img.append(value_line)

        if IP == False:
            plt.pcolormesh(
                self.xpm_xaxis, self.xpm_yaxis, img, cmap="jet", shading="auto"
            )
        elif IP == True:
            ## interpolation
            ip_func = interp2d(self.xpm_xaxis, self.xpm_yaxis, img, kind="linear")
            x_new = np.linspace(
                np.min(self.xpm_xaxis), np.max(self.xpm_xaxis), 10 * len(self.xpm_xaxis)
            )
            y_new = np.linspace(
                np.min(self.xpm_yaxis), np.max(self.xpm_yaxis), 10 * len(self.xpm_yaxis)
            )
            value_new = ip_func(x_new, y_new)
            x_new, y_new = np.meshgrid(x_new, y_new)
            ## show figure
            plt.pcolormesh(x_new, y_new, value_new, cmap="jet", shading="auto")

        ## set ticks and other figure infos
        ax = plt.gca()
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        plt.colorbar(label=self.xpm_legend)
        plt.title(self.xpm_title)
        plt.xlabel(self.xpm_xlabel)
        plt.ylabel(self.xpm_ylabel)
        logging.info("Legend of this xpm figure -> {}".format(self.xpm_legend))

        if outputpng != None:
            plt.savefig(outputpng, dpi=300)
        if noshow == False:
            plt.show()

    def draw_3D(
        self, IP: bool = False, outputpng: str = None, noshow: bool = False
    ) -> None:
        """draw xpm 3D figure (with interpolation)
        IP : whether to interpolation
        outputpng: the name for figure output
        noshow: whether not to show figure, useful for PC without gui
        """

        ## check parameters
        if outputpng != None and os.path.exists(outputpng):
            logging.error("{} already in current directory".format(outputpng))
            sys.exit()

        if self.xpm_type != "Continuous":
            logging.error("Only Continuous type xpm file can draw 3D figure")
            sys.exit()

        ## convert xpm_data to values
        values = []
        for line in self.xpm_datalines:
            for i in range(
                0, self.xpm_width * self.xpm_char_per_pixel, self.xpm_char_per_pixel
            ):
                values.append(
                    float(
                        self.notes[
                            self.chars.index(line[i : i + self.xpm_char_per_pixel])
                        ]
                    )
                )
        xpm_xaxis = np.array(self.xpm_xaxis)
        xpm_yaxis = np.array(self.xpm_yaxis)
        img = np.array(values)

        ## draw 3d figure
        fig = plt.figure()
        plt.clf()
        # ax = fig.gca(projection="3d")
        ax = fig.add_subplot(projection="3d")

        ## interpolation
        IP_value = 1
        if IP == False:
            IP_value = 1
        elif IP == True:
            IP_value = 12
        img = img.reshape(len(xpm_xaxis), len(xpm_yaxis))
        ip_func = interp2d(xpm_xaxis, xpm_yaxis, img, kind="linear")
        x_new = np.linspace(
            np.min(xpm_xaxis), np.max(xpm_xaxis), IP_value * len(xpm_xaxis)
        )
        y_new = np.linspace(
            np.min(xpm_yaxis), np.max(xpm_yaxis), IP_value * len(xpm_yaxis)
        )
        img_new = ip_func(x_new, y_new)
        x_new, y_new = np.meshgrid(x_new, y_new)
        img_new = img_new.reshape(len(x_new), len(y_new))

        ## show figure
        ax.plot_surface(
            x_new,
            y_new,
            img_new,
            alpha=0.9,
            cmap="coolwarm",
            linewidth=0,
            antialiased=False,
        )
        ## set the 2d surface location
        ax.contourf(
            x_new,
            y_new,
            img_new,
            zdir="z",
            offset=math.floor(min(values)) - math.floor(max(values) - min(values)) / 30,
            cmap="coolwarm",
        )

        ## set the axis ticks and other figure infos
        ax.zaxis.set_major_locator(AutoLocator())
        ax.zaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
        # ax.set_zlim(0, )
        # plt.colorbar(surf, shrink=0.6, aspect=12, label=self.xpm_legend)
        plt.title(self.xpm_title)
        ax.set_xlabel(self.xpm_xlabel)
        ax.set_ylabel(self.xpm_ylabel)
        ax.set_zlabel(self.xpm_legend)
        logging.info("Legend of this xpm figure -> {}".format(self.xpm_legend))

        if outputpng != None:
            plt.savefig(outputpng, dpi=300)
        if noshow == False:
            plt.show()


def xpm_combine(
    xpm_files: list = [], outputpng: str = None, noshow: bool = False
) -> None:
    """Combination of xpm files, still in construction, not suggest to use it
    xpm_file_list : a list contains all xpm file names
    outputpng : the name for figure output
    noshow: whether not to show figure, useful for PC without gui
    """

    ###################### bolzmann
    #### G(x) = -kT*Ln(P(x)) + c
    ######################

    x_list, y_list = [], []
    xpm_title, xpm_legend, xpm_xlabel, xpm_ylabel = "", "", "", ""
    for file in xpm_files:
        xpm = XPM(file)
        xpm_title = xpm.xpm_title
        xpm_legend = xpm.xpm_legend
        xpm_xlabel = xpm.xpm_xlabel
        xpm_ylabel = xpm.xpm_ylabel
        if xpm.xpm_type != "Continuous":
            logger.error("can not combine xpm whose type is not Continuous")
            sys.exit()
        scatter_x, scatter_y, _, _, _ = xpm.get_scatter_data()
        x_list += scatter_x
        y_list += scatter_y

    ## combine xpm
    # plt.scatter(x_list, y_list)
    # plt.show()
    heatmap, xedges, yedges = np.histogram2d(x_list, y_list, bins=800)
    heatmap = gaussian_filter(heatmap, sigma=16)
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    # if you set origin="lower", no need to reverse yaxis any more
    plt.imshow(heatmap.T, origin="lower", extent=extent, cmap="jet_r")
    plt.xlim(extent[0], extent[1])
    plt.ylim(extent[2], extent[3])

    ## set ticks and other figure infos
    ax = plt.gca()
    ax.yaxis.set_major_formatter(FormatStrFormatter("%.1f"))
    ax.xaxis.set_major_formatter(FormatStrFormatter("%.1f"))
    plt.title(xpm_title)
    plt.xlabel(xpm_xlabel)
    plt.ylabel(xpm_ylabel)
    logging.info("Legend of this xpm figure -> {}".format(xpm_legend))

    if outputpng != None and os.path.exists(outputpng):
        logging.error("{} already in current directory".format(outputpng))
        sys.exit()
    if outputpng != None:
        plt.savefig(outputpng, dpi=300)
    if noshow == False:
        plt.show()


def xpm_call_functions(arguments: list = None):
    """call functions of xpm module"""
    if arguments == None:
        arguments = [argv for argv in sys.argv]

    ## parse the command parameters
    parser = argparse.ArgumentParser(description="Process xpm files generated by GMX")
    parser.add_argument("-f", "--input", help="input your xpm file")
    parser.add_argument("-o", "--output", help="file name to output")
    parser.add_argument("-x", "--xlabel", type=str, help="the xlabel of figure")
    parser.add_argument("-y", "--ylabel", type=str, help="the ylabel of figure")
    parser.add_argument("-t", "--title", type=str, help="the title of figure")
    parser.add_argument(
        "-xs", "--xshrink", type=str, help="modify x-axis by multipling xshrink"
    )
    parser.add_argument(
        "-ip",
        "--interpolation",
        action="store_true",
        help="whether to apply interpolation (only support Continuous type xpm)",
    )
    parser.add_argument(
        "-pcm",
        "--pcolormesh",
        action="store_true",
        help="whether to apply pcolormesh function to draw",
    )
    parser.add_argument(
        "-3d",
        "--threeDimensions",
        action="store_true",
        help="whether to draw 3D figure",
    )
    parser.add_argument(
        "-ns",
        "--noshow",
        action="store_true",
        help="whether not to show picture, useful on computer without gui",
    )
    parser.add_argument(
        "-c",
        "--combine",
        nargs="+",
        help="specify some xpm files to combine into one figure",
    )

    if len(arguments) < 2:
        logging.error("no input parameters, -h or --help for help messages")
        sys.exit()

    method = arguments[1]
    if method in ["-h", "--help"]:
        parser.parse_args(arguments[1:])
        sys.exit()
    args = parser.parse_args(arguments[2:])

    inputxpm = args.input
    output = args.output
    ip = args.interpolation
    noshow = args.noshow
    pcm = args.pcolormesh
    fig_3d = args.threeDimensions
    xpms2combine = args.combine
    xlabel = args.xlabel
    ylabel = args.ylabel
    title = args.title
    xshrink = args.xshrink

    ## call functions
    if method == "xpm_show":
        if inputxpm == None:
            logging.error("no input file")
            sys.exit()
        xpm = XPM(inputxpm, xlabel, ylabel, title, xshrink)
        if fig_3d == True:
            xpm.draw_3D(ip, output, noshow)
        if pcm == False and fig_3d == False:
            xpm.draw_origin(ip, output, noshow)
        elif pcm == True and fig_3d == False:
            xpm.draw_pcm(ip, output, noshow)
    elif method == "xpm_combine":
        if xpms2combine == None:
            logging.error("no input xpm files for combination")
            sys.exit()
        xpm_combine(xpms2combine, output, noshow)
    elif method == "xpm2csv":
        xpm = XPM(inputxpm, xlabel, ylabel, title, xshrink)
        xpm.xpm2csv(output)
    elif method == "xpm2dat":
        xpm = XPM(inputxpm, xlabel, ylabel, title, xshrink)
        xpm.xpm2dat(output)
    elif method == "xpm2gpl":
        xpm = XPM(inputxpm, xlabel, ylabel, title, xshrink)
        xpm.xpm2gpl(output)
    else:
        logging.error("Wrong method you specified")
        sys.exit()

    logging.info("May you good day !")


def main():
    xpm_call_functions()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(levelname)s -> %(message)s")
    logger = logging.getLogger(__name__)
    main()
